# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import warnings
from collections import OrderedDict
from typing import Callable

import torch
from torch.distributions.kl import kl_divergence, register_kl

import pyro.distributions.torch

from . import constraints
from .distribution import Distribution
from .score_parts import ScoreParts
from .util import broadcast_shape, scale_and_mask


class TorchDistributionMixin(Distribution, Callable):
    """
    Mixin to provide Pyro compatibility for PyTorch distributions.

    You should instead use `TorchDistribution` for new distribution classes.

    This is mainly useful for wrapping existing PyTorch distributions for
    use in Pyro.  Derived classes must first inherit from
    :class:`torch.distributions.distribution.Distribution` and then inherit
    from :class:`TorchDistributionMixin`.
    """

    def __call__(self, sample_shape: torch.Size = torch.Size()) -> torch.Tensor:
        """
        Samples a random value.

        This is reparameterized whenever possible, calling
        :meth:`~torch.distributions.distribution.Distribution.rsample` for
        reparameterized distributions and
        :meth:`~torch.distributions.distribution.Distribution.sample` for
        non-reparameterized distributions.

        :param sample_shape: the size of the iid batch to be drawn from the
            distribution.
        :type sample_shape: torch.Size
        :return: A random value or batch of random values (if parameters are
            batched). The shape of the result should be `self.shape()`.
        :rtype: torch.Tensor
        """
        return (
            self.rsample(sample_shape)
            if self.has_rsample
            else self.sample(sample_shape)
        )

    @property
    def batch_shape(self) -> torch.Size:
        """
        :return: The shape over which parameters are batched.
        :rtype: torch.Size
        """
        raise NotImplementedError

    @property
    def event_shape(self) -> torch.Size:
        """
        :return: The shape of a single sample from the distribution (without batching).
        :rtype: torch.Size
        """
        raise NotImplementedError

    @property
    def event_dim(self) -> int:
        """
        :return: Number of dimensions of individual events.
        :rtype: int
        """
        return len(self.event_shape)

    def shape(self, sample_shape=torch.Size()):
        """
        The tensor shape of samples from this distribution.

        Samples are of shape::

          d.shape(sample_shape) == sample_shape + d.batch_shape + d.event_shape

        :param sample_shape: the size of the iid batch to be drawn from the
            distribution.
        :type sample_shape: torch.Size
        :return: Tensor shape of samples.
        :rtype: torch.Size
        """
        return sample_shape + self.batch_shape + self.event_shape

    @classmethod
    def infer_shapes(cls, **arg_shapes):
        r"""
        Infers ``batch_shape`` and ``event_shape`` given shapes of args to
        :meth:`__init__`.

        .. note:: This assumes distribution shape depends only on the shapes
            of tensor inputs, not in the data contained in those inputs.

        :param \*\*arg_shapes: Keywords mapping name of input arg to
            :class:`torch.Size` or tuple representing the sizes of each
            tensor input.
        :returns: A pair ``(batch_shape, event_shape)`` of the shapes of a
            distribution that would be created with input args of the given
            shapes.
        :rtype: tuple
        """
        if cls.support.event_dim > 0:
            raise NotImplementedError
        # Assumes distribution is univariate.
        batch_shapes = []
        for name, shape in arg_shapes.items():
            event_dim = cls.arg_constraints.get(name, constraints.real).event_dim
            batch_shapes.append(shape[: len(shape) - event_dim])
        batch_shape = torch.Size(broadcast_shape(*batch_shapes))
        event_shape = torch.Size()
        return batch_shape, event_shape

    def expand(self, batch_shape, _instance=None) -> "ExpandedDistribution":
        """
        Returns a new :class:`ExpandedDistribution` instance with batch
        dimensions expanded to `batch_shape`.

        :param tuple batch_shape: batch shape to expand to.
        :param _instance: unused argument for compatibility with
            :meth:`torch.distributions.Distribution.expand`
        :return: an instance of `ExpandedDistribution`.
        :rtype: :class:`ExpandedDistribution`
        """
        return ExpandedDistribution(self, batch_shape)

    def expand_by(self, sample_shape):
        """
        Expands a distribution by adding ``sample_shape`` to the left side of
        its :attr:`~torch.distributions.distribution.Distribution.batch_shape`.

        To expand internal dims of ``self.batch_shape`` from 1 to something
        larger, use :meth:`expand` instead.

        :param torch.Size sample_shape: The size of the iid batch to be drawn
            from the distribution.
        :return: An expanded version of this distribution.
        :rtype: :class:`ExpandedDistribution`
        """
        try:
            expanded_dist = self.expand(torch.Size(sample_shape) + self.batch_shape)
        except NotImplementedError:
            expanded_dist = TorchDistributionMixin.expand(
                self, torch.Size(sample_shape) + self.batch_shape
            )
        return expanded_dist

    def reshape(self, sample_shape=None, extra_event_dims=None):
        raise Exception(
            """
            .reshape(sample_shape=s, extra_event_dims=n) was renamed and split into
            .expand_by(sample_shape=s).to_event(reinterpreted_batch_ndims=n)."""
        )

    def to_event(self, reinterpreted_batch_ndims=None):
        """
        Reinterprets the ``n`` rightmost dimensions of this distributions
        :attr:`~torch.distributions.distribution.Distribution.batch_shape`
        as event dims, adding them to the left side of
        :attr:`~torch.distributions.distribution.Distribution.event_shape`.

        Example:

            .. doctest::
               :hide:

               >>> d0 = dist.Normal(torch.zeros(2, 3, 4, 5), torch.ones(2, 3, 4, 5))
               >>> [d0.batch_shape, d0.event_shape]
               [torch.Size([2, 3, 4, 5]), torch.Size([])]
               >>> d1 = d0.to_event(2)

            >>> [d1.batch_shape, d1.event_shape]
            [torch.Size([2, 3]), torch.Size([4, 5])]
            >>> d2 = d1.to_event(1)
            >>> [d2.batch_shape, d2.event_shape]
            [torch.Size([2]), torch.Size([3, 4, 5])]
            >>> d3 = d1.to_event(2)
            >>> [d3.batch_shape, d3.event_shape]
            [torch.Size([]), torch.Size([2, 3, 4, 5])]

        :param int reinterpreted_batch_ndims: The number of batch dimensions to
            reinterpret as event dimensions. May be negative to remove
            dimensions from an :class:`pyro.distributions.torch.Independent` .
            If None, convert all dimensions to event dimensions.
        :return: A reshaped version of this distribution.
        :rtype: :class:`pyro.distributions.torch.Independent`
        """
        if reinterpreted_batch_ndims is None:
            reinterpreted_batch_ndims = len(self.batch_shape)

        # Deconstruct Independent distributions.
        base_dist = self
        while isinstance(base_dist, torch.distributions.Independent):
            reinterpreted_batch_ndims += base_dist.reinterpreted_batch_ndims
            base_dist = base_dist.base_dist

        if reinterpreted_batch_ndims == 0:
            return base_dist
        if reinterpreted_batch_ndims < 0:
            raise ValueError(
                "Cannot remove event dimensions from {}".format(type(self))
            )
        return pyro.distributions.torch.Independent(
            base_dist, reinterpreted_batch_ndims
        )

    def independent(self, reinterpreted_batch_ndims=None):
        warnings.warn(
            "independent is deprecated; use to_event instead", DeprecationWarning
        )
        return self.to_event(reinterpreted_batch_ndims=reinterpreted_batch_ndims)

    def mask(self, mask):
        """
        Masks a distribution by a boolean or boolean-valued tensor that is
        broadcastable to the distributions
        :attr:`~torch.distributions.distribution.Distribution.batch_shape` .

        :param mask: A boolean or boolean valued tensor.
        :type mask: bool or torch.Tensor
        :return: A masked copy of this distribution.
        :rtype: :class:`MaskedDistribution`
        """
        return MaskedDistribution(self, mask)


class TorchDistribution(torch.distributions.Distribution, TorchDistributionMixin):
    """
    Base class for PyTorch-compatible distributions with Pyro support.

    This should be the base class for almost all new Pyro distributions.

    .. note::

        Parameters and data should be of type :class:`~torch.Tensor`
        and all methods return type :class:`~torch.Tensor` unless
        otherwise noted.

    **Tensor Shapes**:

    TorchDistributions provide a method ``.shape()`` for the tensor shape of samples::

      x = d.sample(sample_shape)
      assert x.shape == d.shape(sample_shape)

    Pyro follows the same distribution shape semantics as PyTorch. It distinguishes
    between three different roles for tensor shapes of samples:

    - *sample shape* corresponds to the shape of the iid samples drawn from the distribution.
      This is taken as an argument by the distribution's `sample` method.
    - *batch shape* corresponds to non-identical (independent) parameterizations of
      the distribution, inferred from the distribution's parameter shapes. This is
      fixed for a distribution instance.
    - *event shape* corresponds to the event dimensions of the distribution, which
      is fixed for a distribution class. These are collapsed when we try to score
      a sample from the distribution via `d.log_prob(x)`.

    These shapes are related by the equation::

      assert d.shape(sample_shape) == sample_shape + d.batch_shape + d.event_shape

    Distributions provide a vectorized
    :meth:`~torch.distributions.distribution.Distribution.log_prob` method that
    evaluates the log probability density of each event in a batch
    independently, returning a tensor of shape
    ``sample_shape + d.batch_shape``::

      x = d.sample(sample_shape)
      assert x.shape == d.shape(sample_shape)
      log_p = d.log_prob(x)
      assert log_p.shape == sample_shape + d.batch_shape

    **Implementing New Distributions**:

    Derived classes must implement the methods
    :meth:`~torch.distributions.distribution.Distribution.sample`
    (or :meth:`~torch.distributions.distribution.Distribution.rsample` if
    ``.has_rsample == True``) and
    :meth:`~torch.distributions.distribution.Distribution.log_prob`, and must
    implement the properties
    :attr:`~torch.distributions.distribution.Distribution.batch_shape`,
    and :attr:`~torch.distributions.distribution.Distribution.event_shape`.
    Discrete classes may also implement the
    :meth:`~torch.distributions.distribution.Distribution.enumerate_support`
    method to improve gradient estimates and set
    ``.has_enumerate_support = True``.
    """

    # Provides a default `.expand` method for Pyro distributions which overrides
    # torch.distributions.Distribution.expand (throws a NotImplementedError).
    expand = TorchDistributionMixin.expand


class MaskedDistribution(TorchDistribution):
    """
    Masks a distribution by a boolean tensor that is broadcastable to the
    distribution's :attr:`~torch.distributions.distribution.Distribution.batch_shape`.

    In the special case ``mask is False``, computation of :meth:`log_prob` ,
    :meth:`score_parts` , and ``kl_divergence()`` is skipped, and constant zero
    values are returned instead.

    :param mask: A boolean or boolean-valued tensor.
    :type mask: torch.Tensor or bool
    """

    arg_constraints = {}

    def __init__(self, base_dist, mask):
        if isinstance(mask, bool):
            self._mask = mask
        else:
            batch_shape = broadcast_shape(mask.shape, base_dist.batch_shape)
            if mask.shape != batch_shape:
                mask = mask.expand(batch_shape)
            if base_dist.batch_shape != batch_shape:
                base_dist = base_dist.expand(batch_shape)
            self._mask = mask.bool()
        self.base_dist = base_dist
        super().__init__(base_dist.batch_shape, base_dist.event_shape)

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(MaskedDistribution, _instance)
        batch_shape = torch.Size(batch_shape)
        new.base_dist = self.base_dist.expand(batch_shape)
        new._mask = self._mask
        if isinstance(new._mask, torch.Tensor):
            new._mask = new._mask.expand(batch_shape)
        super(MaskedDistribution, new).__init__(
            batch_shape, self.event_shape, validate_args=False
        )
        new._validate_args = self._validate_args
        return new

    @property
    def has_rsample(self):
        return self.base_dist.has_rsample

    @property
    def has_enumerate_support(self):
        return self.base_dist.has_enumerate_support

    @constraints.dependent_property
    def support(self):
        return self.base_dist.support

    def sample(self, sample_shape=torch.Size()):
        return self.base_dist.sample(sample_shape)

    def rsample(self, sample_shape=torch.Size()):
        return self.base_dist.rsample(sample_shape)

    def log_prob(self, value):
        if self._mask is False:
            shape = broadcast_shape(
                self.base_dist.batch_shape, value.shape[: value.dim() - self.event_dim]
            )
            return torch.zeros((), device=value.device).expand(shape)
        if self._mask is True:
            return self.base_dist.log_prob(value)
        return scale_and_mask(self.base_dist.log_prob(value), mask=self._mask)

    def score_parts(self, value):
        if isinstance(self._mask, bool):
            return super().score_parts(value)  # calls self.log_prob(value)
        return self.base_dist.score_parts(value).scale_and_mask(mask=self._mask)

    def enumerate_support(self, expand=True):
        return self.base_dist.enumerate_support(expand=expand)

    @property
    def mean(self):
        return self.base_dist.mean

    @property
    def variance(self):
        return self.base_dist.variance

    def conjugate_update(self, other):
        """
        EXPERIMENTAL.
        """
        updated, log_normalizer = self.base_dist.conjugate_update(other)
        updated = updated.mask(self._mask)
        log_normalizer = torch.where(
            self._mask, log_normalizer, torch.zeros_like(log_normalizer)
        )
        return updated, log_normalizer


class ExpandedDistribution(TorchDistribution):
    arg_constraints = {}

    def __init__(self, base_dist, batch_shape=torch.Size()):
        self.base_dist = base_dist
        super().__init__(base_dist.batch_shape, base_dist.event_shape)
        # adjust batch shape
        self.expand(batch_shape)

    def expand(self, batch_shape, _instance=None):
        # Do basic validation. e.g. we should not "unexpand" distributions even if that is possible.
        new_shape, _, _ = self._broadcast_shape(self.batch_shape, batch_shape)
        # Record interstitial and expanded dims/sizes w.r.t. the base distribution
        new_shape, expanded_sizes, interstitial_sizes = self._broadcast_shape(
            self.base_dist.batch_shape, new_shape
        )
        self._batch_shape = new_shape
        self._expanded_sizes = expanded_sizes
        self._interstitial_sizes = interstitial_sizes
        return self

    @staticmethod
    def _broadcast_shape(existing_shape, new_shape):
        if len(new_shape) < len(existing_shape):
            raise ValueError(
                "Cannot broadcast distribution of shape {} to shape {}".format(
                    existing_shape, new_shape
                )
            )
        reversed_shape = list(reversed(existing_shape))
        expanded_sizes, interstitial_sizes = [], []
        for i, size in enumerate(reversed(new_shape)):
            if i >= len(reversed_shape):
                reversed_shape.append(size)
                expanded_sizes.append((-i - 1, size))
            elif reversed_shape[i] == 1:
                if size != 1:
                    reversed_shape[i] = size
                    interstitial_sizes.append((-i - 1, size))
            elif reversed_shape[i] != size:
                raise ValueError(
                    "Cannot broadcast distribution of shape {} to shape {}".format(
                        existing_shape, new_shape
                    )
                )
        return (
            tuple(reversed(reversed_shape)),
            OrderedDict(expanded_sizes),
            OrderedDict(interstitial_sizes),
        )

    @property
    def has_rsample(self):
        return self.base_dist.has_rsample

    @property
    def has_enumerate_support(self):
        return self.base_dist.has_enumerate_support

    @constraints.dependent_property
    def support(self):
        return self.base_dist.support

    def _sample(self, sample_fn, sample_shape):
        interstitial_dims = tuple(self._interstitial_sizes.keys())
        interstitial_dims = tuple(i - self.event_dim for i in interstitial_dims)
        interstitial_sizes = tuple(self._interstitial_sizes.values())
        expanded_sizes = tuple(self._expanded_sizes.values())
        batch_shape = expanded_sizes + interstitial_sizes
        samples = sample_fn(sample_shape + batch_shape)
        interstitial_idx = len(sample_shape) + len(expanded_sizes)
        interstitial_sample_dims = tuple(
            range(interstitial_idx, interstitial_idx + len(interstitial_sizes))
        )
        for dim1, dim2 in zip(interstitial_dims, interstitial_sample_dims):
            samples = samples.transpose(dim1, dim2)
        return samples.reshape(sample_shape + self.batch_shape + self.event_shape)

    def sample(self, sample_shape=torch.Size()):
        return self._sample(self.base_dist.sample, sample_shape)

    def rsample(self, sample_shape=torch.Size()):
        return self._sample(self.base_dist.rsample, sample_shape)

    def log_prob(self, value):
        shape = broadcast_shape(
            self.batch_shape, value.shape[: value.dim() - self.event_dim]
        )
        log_prob = self.base_dist.log_prob(value)
        return log_prob.expand(shape)

    def score_parts(self, value):
        shape = broadcast_shape(
            self.batch_shape, value.shape[: value.dim() - self.event_dim]
        )
        log_prob, score_function, entropy_term = self.base_dist.score_parts(value)
        if self.batch_shape != self.base_dist.batch_shape:
            log_prob = log_prob.expand(shape)
            if isinstance(score_function, torch.Tensor):
                score_function = score_function.expand(shape)
            if isinstance(entropy_term, torch.Tensor):
                entropy_term = entropy_term.expand(shape)
        return ScoreParts(log_prob, score_function, entropy_term)

    def enumerate_support(self, expand=True):
        samples = self.base_dist.enumerate_support(expand=False)
        enum_shape = samples.shape[:1]
        samples = samples.reshape(enum_shape + (1,) * len(self.batch_shape))
        if expand:
            samples = samples.expand(enum_shape + self.batch_shape)
        return samples

    @property
    def mean(self):
        return self.base_dist.mean.expand(self.batch_shape + self.event_shape)

    @property
    def variance(self):
        return self.base_dist.variance.expand(self.batch_shape + self.event_shape)

    def conjugate_update(self, other):
        """
        EXPERIMENTAL.
        """
        updated, log_normalizer = self.base_dist.conjugate_update(other)
        updated = updated.expand(self.batch_shape)
        log_normalizer = log_normalizer.expand(self.batch_shape)
        return updated, log_normalizer


@register_kl(MaskedDistribution, MaskedDistribution)
def _kl_masked_masked(p, q):
    if p._mask is False or q._mask is False:
        mask = False
    elif p._mask is True:
        mask = q._mask
    elif q._mask is True:
        mask = p._mask
    elif p._mask is q._mask:
        mask = p._mask
    else:
        mask = p._mask & q._mask

    if mask is False:
        return 0.0  # Return a float, since we cannot determine device.
    if mask is True:
        return kl_divergence(p.base_dist, q.base_dist)
    kl = kl_divergence(p.base_dist, q.base_dist)
    return scale_and_mask(kl, mask=mask)



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import torch

from pyro.contrib.gp.parameterized import Parameterized


def _zero_mean_function(x):
    return 0


class GPModel(Parameterized):
    r"""
    Base class for Gaussian Process models.

    The core of a Gaussian Process is a covariance function :math:`k` which governs
    the similarity between input points. Given :math:`k`, we can establish a
    distribution over functions :math:`f` by a multivarite normal distribution

    .. math:: p(f(X)) = \mathcal{N}(0, k(X, X)),

    where :math:`X` is any set of input points and :math:`k(X, X)` is a covariance
    matrix whose entries are outputs :math:`k(x, z)` of :math:`k` over input pairs
    :math:`(x, z)`. This distribution is usually denoted by

    .. math:: f \sim \mathcal{GP}(0, k).

    .. note:: Generally, beside a covariance matrix :math:`k`, a Gaussian Process can
        also be specified by a mean function :math:`m` (which is a zero-value function
        by default). In that case, its distribution will be

        .. math:: p(f(X)) = \mathcal{N}(m(X), k(X, X)).

    Gaussian Process models are :class:`~pyro.contrib.gp.util.Parameterized`
    subclasses. So its parameters can be learned, set priors, or fixed by using
    corresponding methods from :class:`~pyro.contrib.gp.util.Parameterized`. A typical
    way to define a Gaussian Process model is

        >>> X = torch.tensor([[1., 5, 3], [4, 3, 7]])
        >>> y = torch.tensor([2., 1])
        >>> kernel = gp.kernels.RBF(input_dim=3)
        >>> kernel.variance = pyro.nn.PyroSample(dist.Uniform(torch.tensor(0.5), torch.tensor(1.5)))
        >>> kernel.lengthscale = pyro.nn.PyroSample(dist.Uniform(torch.tensor(1.0), torch.tensor(3.0)))
        >>> gpr = gp.models.GPRegression(X, y, kernel)

    There are two ways to train a Gaussian Process model:

    + Using an MCMC algorithm (in module :mod:`pyro.infer.mcmc`) on :meth:`model` to
      get posterior samples for the Gaussian Process's parameters. For example:

        >>> hmc_kernel = HMC(gpr.model)
        >>> mcmc = MCMC(hmc_kernel, num_samples=10)
        >>> mcmc.run()
        >>> ls_name = "kernel.lengthscale"
        >>> posterior_ls = mcmc.get_samples()[ls_name]

    + Using a variational inference on the pair :meth:`model`, :meth:`guide`:

        >>> optimizer = torch.optim.Adam(gpr.parameters(), lr=0.01)
        >>> loss_fn = pyro.infer.TraceMeanField_ELBO().differentiable_loss
        >>>
        >>> for i in range(1000):
        ...     optimizer.zero_grad()
        ...     loss = loss_fn(gpr.model, gpr.guide)  # doctest: +SKIP
        ...     loss.backward()  # doctest: +SKIP
        ...     optimizer.step()

    To give a prediction on new dataset, simply use :meth:`forward` like any PyTorch
    :class:`torch.nn.Module`:

        >>> Xnew = torch.tensor([[2., 3, 1]])
        >>> f_loc, f_cov = gpr(Xnew, full_cov=True)

    Reference:

    [1] `Gaussian Processes for Machine Learning`,
    Carl E. Rasmussen, Christopher K. I. Williams

    :param torch.Tensor X: A input data for training. Its first dimension is the number
        of data points.
    :param torch.Tensor y: An output data for training. Its last dimension is the
        number of data points.
    :param ~pyro.contrib.gp.kernels.kernel.Kernel kernel: A Pyro kernel object, which
        is the covariance function :math:`k`.
    :param callable mean_function: An optional mean function :math:`m` of this Gaussian
        process. By default, we use zero mean.
    :param float jitter: A small positive term which is added into the diagonal part of
        a covariance matrix to help stablize its Cholesky decomposition.
    """

    def __init__(self, X, y, kernel, mean_function=None, jitter=1e-6):
        assert isinstance(
            X, torch.Tensor
        ), "X needs to be a torch Tensor instead of a {}".format(type(X))
        if y is not None:
            assert isinstance(
                y, torch.Tensor
            ), "y needs to be a torch Tensor instead of a {}".format(type(y))

        super().__init__()
        self.set_data(X, y)
        self.kernel = kernel
        self.mean_function = (
            mean_function if mean_function is not None else _zero_mean_function
        )
        self.jitter = jitter

    def model(self):
        """
        A "model" stochastic function. If ``self.y`` is ``None``, this method returns
        mean and variance of the Gaussian Process prior.
        """
        raise NotImplementedError

    def guide(self):
        """
        A "guide" stochastic function to be used in variational inference methods. It
        also gives posterior information to the method :meth:`forward` for prediction.
        """
        raise NotImplementedError

    def forward(self, Xnew, full_cov=False):
        r"""
        Computes the mean and covariance matrix (or variance) of Gaussian Process
        posterior on a test input data :math:`X_{new}`:

        .. math:: p(f^* \mid X_{new}, X, y, k, \theta),

        where :math:`\theta` are parameters of this model.

        .. note:: Model's parameters :math:`\theta` together with kernel's parameters
            have been learned from a training procedure (MCMC or SVI).

        :param torch.Tensor Xnew: A input data for testing. Note that
            ``Xnew.shape[1:]`` must be the same as ``X.shape[1:]``.
        :param bool full_cov: A flag to decide if we want to predict full covariance
            matrix or just variance.
        :returns: loc and covariance matrix (or variance) of :math:`p(f^*(X_{new}))`
        :rtype: tuple(torch.Tensor, torch.Tensor)
        """
        raise NotImplementedError

    def set_data(self, X, y=None):
        """
        Sets data for Gaussian Process models.

        Some examples to utilize this method are:

        .. doctest::
           :hide:

            >>> X = torch.tensor([[1., 5, 3], [4, 3, 7]])
            >>> y = torch.tensor([2., 1])
            >>> kernel = gp.kernels.RBF(input_dim=3)
            >>> kernel.variance = pyro.nn.PyroSample(dist.Uniform(torch.tensor(0.5), torch.tensor(1.5)))
            >>> kernel.lengthscale = pyro.nn.PyroSample(dist.Uniform(torch.tensor(1.0), torch.tensor(3.0)))

        + Batch training on a sparse variational model:

            >>> Xu = torch.tensor([[1., 0, 2]])  # inducing input
            >>> likelihood = gp.likelihoods.Gaussian()
            >>> vsgp = gp.models.VariationalSparseGP(X, y, kernel, Xu, likelihood)
            >>> optimizer = torch.optim.Adam(vsgp.parameters(), lr=0.01)
            >>> loss_fn = pyro.infer.TraceMeanField_ELBO().differentiable_loss
            >>> batched_X, batched_y = X.split(split_size=10), y.split(split_size=10)
            >>> for Xi, yi in zip(batched_X, batched_y):
            ...     optimizer.zero_grad()
            ...     vsgp.set_data(Xi, yi)
            ...     loss = loss_fn(vsgp.model, vsgp.guide)  # doctest: +SKIP
            ...     loss.backward()  # doctest: +SKIP
            ...     optimizer.step()

        + Making a two-layer Gaussian Process stochastic function:

            >>> gpr1 = gp.models.GPRegression(X, None, kernel)
            >>> Z, _ = gpr1.model()
            >>> gpr2 = gp.models.GPRegression(Z, y, kernel)
            >>> def two_layer_model():
            ...     Z, _ = gpr1.model()
            ...     gpr2.set_data(Z, y)
            ...     return gpr2.model()

        References:

        [1] `Scalable Variational Gaussian Process Classification`,
        James Hensman, Alexander G. de G. Matthews, Zoubin Ghahramani

        [2] `Deep Gaussian Processes`,
        Andreas C. Damianou, Neil D. Lawrence

        :param torch.Tensor X: A input data for training. Its first dimension is the
            number of data points.
        :param torch.Tensor y: An output data for training. Its last dimension is the
            number of data points.
        """
        if y is not None and X.size(0) != y.size(-1):
            raise ValueError(
                "Expected the number of input data points equal to the "
                "number of output data points, but got {} and {}.".format(
                    X.size(0), y.size(-1)
                )
            )
        self.X = X
        self.y = y

    def _check_Xnew_shape(self, Xnew):
        """
        Checks the correction of the shape of new data.

        :param torch.Tensor Xnew: A input data for testing. Note that
            ``Xnew.shape[1:]`` must be the same as ``self.X.shape[1:]``.
        """
        if Xnew.dim() != self.X.dim():
            raise ValueError(
                "Train data and test data should have the same "
                "number of dimensions, but got {} and {}.".format(
                    self.X.dim(), Xnew.dim()
                )
            )
        if self.X.shape[1:] != Xnew.shape[1:]:
            raise ValueError(
                "Train data and test data should have the same "
                "shape of features, but got {} and {}.".format(
                    self.X.shape[1:], Xnew.shape[1:]
                )
            )



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

from typing import Optional, Tuple

import torch

from pyro.distributions.distribution import Distribution
from pyro.poutine.indep_messenger import CondIndepStackFrame, IndepMessenger
from pyro.poutine.runtime import Message, apply_stack
from pyro.poutine.util import is_validation_enabled
from pyro.util import ignore_jit_warnings


class _Subsample(Distribution):
    """
    Randomly select a subsample of a range of indices.

    Internal use only. This should only be used by `plate`.
    """

    def __init__(
        self,
        size: int,
        subsample_size: Optional[int],
        use_cuda: Optional[bool] = None,
        device: Optional[str] = None,
    ) -> None:
        """
        :param int size: the size of the range to subsample from
        :param int subsample_size: the size of the returned subsample
        :param bool use_cuda: DEPRECATED, use the `device` arg instead.
            Whether to use cuda tensors.
        :param str device: device to place the `sample` and `log_prob`
            results on.
        """
        self.size = size
        self.subsample_size = subsample_size
        self.use_cuda = use_cuda
        if self.use_cuda is not None:
            if self.use_cuda ^ (device != "cpu"):
                raise ValueError(
                    "Incompatible arg values use_cuda={}, device={}.".format(
                        use_cuda, device
                    )
                )
        with ignore_jit_warnings(["torch.Tensor results are registered as constants"]):
            self.device = device or torch.Tensor().device

    @ignore_jit_warnings(["Converting a tensor to a Python boolean"])
    def sample(self, sample_shape: torch.Size = torch.Size()) -> torch.Tensor:
        """
        :returns: a random subsample of `range(size)`
        :rtype: torch.LongTensor
        """
        if sample_shape:
            raise NotImplementedError
        subsample_size = self.subsample_size
        if subsample_size is None or subsample_size >= self.size:
            result = torch.arange(self.size, device=self.device)
        else:
            result = torch.randperm(self.size, device=self.device)[
                :subsample_size
            ].clone()
        return result.cuda() if self.use_cuda else result

    def log_prob(self, x: torch.Tensor) -> torch.Tensor:
        # This is zero so that plate can provide an unbiased estimate of
        # the non-subsampled log_prob.
        result = torch.tensor(0.0, device=self.device)
        return result.cuda() if self.use_cuda else result


class SubsampleMessenger(IndepMessenger):
    """
    Extension of IndepMessenger that includes subsampling.
    """

    def __init__(
        self,
        name: str,
        size: Optional[int] = None,
        subsample_size: Optional[int] = None,
        subsample: Optional[torch.Tensor] = None,
        dim: Optional[int] = None,
        use_cuda: Optional[bool] = None,
        device: Optional[str] = None,
    ) -> None:
        full_size, self.subsample_size, subsample = self._subsample(
            name,
            size,
            subsample_size,
            subsample,
            use_cuda,
            device,
        )
        super().__init__(name, full_size, dim, device)
        self._indices = subsample

    @staticmethod
    def _subsample(
        name: str,
        size: Optional[int] = None,
        subsample_size: Optional[int] = None,
        subsample: Optional[torch.Tensor] = None,
        use_cuda: Optional[bool] = None,
        device: Optional[str] = None,
    ) -> Tuple[int, int, Optional[torch.Tensor]]:
        """
        Helper function for plate. See its docstrings for details.
        """
        if size is None:
            assert subsample_size is None
            assert subsample is None
            size = -1  # This is PyTorch convention for "arbitrary size"
            subsample_size = -1
        else:
            msg = Message(
                type="sample",
                name=name,
                fn=_Subsample(size, subsample_size, use_cuda, device),
                is_observed=False,
                args=(),
                kwargs={},
                value=subsample,
                infer={},
                scale=1.0,
                mask=None,
                cond_indep_stack=(),
                done=False,
                stop=False,
                continuation=None,
            )
            apply_stack(msg)
            subsample = msg["value"]

        with ignore_jit_warnings():
            if subsample_size is None:
                assert subsample is not None
                subsample_size = (
                    subsample.size(0)
                    if isinstance(subsample, torch.Tensor)
                    else len(subsample)
                )
            elif subsample is not None and subsample_size != len(subsample):
                raise ValueError(
                    "subsample_size does not match len(subsample), {} vs {}.".format(
                        subsample_size, len(subsample)
                    )
                    + " Did you accidentally use different subsample_size in the model and guide?"
                )

        return size, subsample_size, subsample

    def _reset(self) -> None:
        self._indices = None
        super()._reset()

    def _process_message(self, msg: Message) -> None:
        frame = CondIndepStackFrame(
            name=self.name,
            dim=self.dim,
            size=self.subsample_size,
            counter=self.counter,
            full_size=self.size,  # used for param initialization
        )
        msg["cond_indep_stack"] = (frame,) + msg["cond_indep_stack"]
        if isinstance(self.size, torch.Tensor) or isinstance(  # type: ignore[unreachable]
            self.subsample_size, torch.Tensor  # type: ignore[unreachable]
        ):
            if not isinstance(msg["scale"], torch.Tensor):  # type: ignore[unreachable]
                with ignore_jit_warnings():
                    msg["scale"] = torch.tensor(msg["scale"])
        msg["scale"] = msg["scale"] * self.size / self.subsample_size

    def _postprocess_message(self, msg: Message) -> None:
        if msg["type"] in ("param", "subsample") and self.dim is not None:
            event_dim = msg["kwargs"].get("event_dim")
            if event_dim is not None:
                assert event_dim >= 0
                dim = self.dim - event_dim
                assert msg["value"] is not None
                shape = msg["value"].shape
                if len(shape) >= -dim and shape[dim] != 1:
                    if is_validation_enabled() and shape[dim] != self.size:
                        if msg["type"] == "param":
                            statement = "pyro.param({}, ..., event_dim={})".format(
                                msg["name"], event_dim
                            )
                        else:
                            statement = "pyro.subsample(..., event_dim={})".format(
                                event_dim
                            )
                        raise ValueError(
                            "Inside pyro.plate({}, {}, dim={}) invalid shape of {}: {}".format(
                                self.name, self.size, self.dim, statement, shape
                            )
                        )
                    # Subsample parameters with known batch semantics.
                    if self.subsample_size < self.size:
                        value = msg["value"]
                        assert self._indices is not None
                        new_value = value.index_select(
                            dim, self._indices.to(value.device)
                        )
                        if msg["type"] == "param":
                            if hasattr(value, "_pyro_unconstrained_param"):
                                param = value._pyro_unconstrained_param  # type: ignore[attr-defined]
                            else:
                                param = value.unconstrained()  # type: ignore[attr-defined]

                            if not hasattr(param, "_pyro_subsample"):
                                param._pyro_subsample = {}

                            param._pyro_subsample[dim] = self._indices
                            new_value._pyro_unconstrained_param = param  # type: ignore[attr-defined]
                        msg["value"] = new_value



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import math

import torch
import torch.nn.functional as F
from torch.distributions.transforms import TanhTransform, Transform

from .. import constraints

# TODO: Move upstream


class ELUTransform(Transform):
    r"""
    Bijective transform via the mapping :math:`y = \text{ELU}(x)`.
    """

    domain = constraints.real
    codomain = constraints.positive
    bijective = True
    sign = +1

    def __eq__(self, other):
        return isinstance(other, ELUTransform)

    def _call(self, x):
        return F.elu(x)

    def _inverse(self, y, eps=1e-8):
        return torch.max(y, torch.zeros_like(y)) + torch.min(
            torch.log1p(y + eps), torch.zeros_like(y)
        )

    def log_abs_det_jacobian(self, x, y):
        return -F.relu(-x)


def elu():
    """
    A helper function to create an
    :class:`~pyro.distributions.transform.ELUTransform` object for consistency with
    other helpers.
    """
    return ELUTransform()


# TODO: Move upstream


class LeakyReLUTransform(Transform):
    r"""
    Bijective transform via the mapping :math:`y = \text{LeakyReLU}(x)`.
    """

    domain = constraints.real
    codomain = constraints.real
    bijective = True
    sign = +1

    def __eq__(self, other):
        return isinstance(other, LeakyReLUTransform)

    def _call(self, x):
        return F.leaky_relu(x)

    def _inverse(self, y):
        return F.leaky_relu(y, negative_slope=100.0)

    def log_abs_det_jacobian(self, x, y):
        return torch.where(
            x >= 0.0, torch.zeros_like(x), torch.ones_like(x) * math.log(0.01)
        )


def leaky_relu():
    """
    A helper function to create a
    :class:`~pyro.distributions.transforms.LeakyReLUTransform` object for
    consistency with other helpers.
    """
    return LeakyReLUTransform()


def tanh():
    """
    A helper function to create a
    :class:`~pyro.distributions.transforms.TanhTransform` object for consistency
    with other helpers.
    """
    return TanhTransform()



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import warnings
from collections import OrderedDict
from functools import partial

from torch.distributions import biject_to, constraints
from torch.nn import Parameter

import pyro
import pyro.distributions as dist
from pyro.distributions.util import eye_like
from pyro.nn.module import PyroModule, PyroParam, PyroSample, pyro_method


def _is_real_support(support):
    if isinstance(support, pyro.distributions.constraints.independent):
        return _is_real_support(support.base_constraint)
    else:
        return support in [constraints.real, constraints.real_vector]


def _get_sample_fn(module, name):
    if module.mode == "model":
        return module._priors[name]

    dist_constructor, dist_args = module._guides[name]

    if dist_constructor is dist.Delta:
        p_map = getattr(module, "{}_map".format(name))
        return dist.Delta(p_map, event_dim=p_map.dim())

    # create guide
    dist_args = {arg: getattr(module, "{}_{}".format(name, arg)) for arg in dist_args}
    guide = dist_constructor(**dist_args)

    # no need to do transforms when support is real (for mean field ELBO)
    support = module._priors[name].support
    if _is_real_support(support):
        return guide.to_event()

    # otherwise, we do inference in unconstrained space and transform the value
    # back to original space
    # TODO: move this logic to infer.autoguide or somewhere else
    unconstrained_value = pyro.sample(
        module._pyro_get_fullname("{}_latent".format(name)),
        guide.to_event(),
        infer={"is_auxiliary": True},
    )
    transform = biject_to(support)
    value = transform(unconstrained_value)
    log_density = transform.inv.log_abs_det_jacobian(value, unconstrained_value)
    return dist.Delta(value, log_density.sum(), event_dim=value.dim())


class Parameterized(PyroModule):
    """
    A wrapper of :class:`~pyro.nn.module.PyroModule` whose parameters can be set
    constraints, set priors.

    By default, when we set a prior to a parameter, an auto Delta guide will be
    created. We can use the method :meth:`autoguide` to setup other auto guides.

    Example::

        >>> class Linear(Parameterized):
        ...     def __init__(self, a, b):
        ...         super().__init__()
        ...         self.a = Parameter(a)
        ...         self.b = Parameter(b)
        ...
        ...     def forward(self, x):
        ...         return self.a * x + self.b
        ...
        >>> linear = Linear(torch.tensor(1.), torch.tensor(0.))
        >>> linear.a = PyroParam(torch.tensor(1.), constraints.positive)
        >>> linear.b = PyroSample(dist.Normal(0, 1))
        >>> linear.autoguide("b", dist.Normal)
        >>> assert "a_unconstrained" in dict(linear.named_parameters())
        >>> assert "b_loc" in dict(linear.named_parameters())
        >>> assert "b_scale_unconstrained" in dict(linear.named_parameters())

    Note that by default, data of a parameter is a float :class:`torch.Tensor`
    (unless we use :func:`torch.set_default_dtype` to change default
    tensor type). To cast these parameters to a correct data type or GPU device,
    we can call methods such as :meth:`~torch.nn.Module.double` or
    :meth:`~torch.nn.Module.cuda`. See :class:`torch.nn.Module` for more
    information.
    """

    def __init__(self):
        super().__init__()
        self._priors = OrderedDict()
        self._guides = OrderedDict()
        self._mode = "model"

    def set_prior(self, name, prior):
        """
        Sets prior for a parameter.

        :param str name: Name of the parameter.
        :param ~pyro.distributions.distribution.Distribution prior: A Pyro prior
            distribution.
        """
        warnings.warn(
            "The method `self.set_prior({}, prior)` has been deprecated"
            " in favor of `self.{} = PyroSample(prior)`.".format(name, name),
            UserWarning,
        )
        setattr(self, name, PyroSample(prior))

    def __setattr__(self, name, value):
        if isinstance(value, PyroSample):
            prior = value.prior
            if hasattr(prior, "sample"):
                self._priors[name] = prior
                self.autoguide(name, dist.Delta)
                value = PyroSample(partial(_get_sample_fn, name=name))
        super().__setattr__(name, value)

    def autoguide(self, name, dist_constructor):
        """
        Sets an autoguide for an existing parameter with name ``name`` (mimic
        the behavior of module :mod:`pyro.infer.autoguide`).

        .. note:: `dist_constructor` should be one of
            :class:`~pyro.distributions.Delta`,
            :class:`~pyro.distributions.Normal`, and
            :class:`~pyro.distributions.MultivariateNormal`. More distribution
            constructor will be supported in the future if needed.

        :param str name: Name of the parameter.
        :param dist_constructor: A
            :class:`~pyro.distributions.distribution.Distribution` constructor.
        """
        if name not in self._priors:
            raise ValueError("There is no prior for parameter: {}".format(name))

        if dist_constructor not in [dist.Delta, dist.Normal, dist.MultivariateNormal]:
            raise NotImplementedError(
                "Unsupported distribution type: {}".format(dist_constructor)
            )

        # delete old guide
        if name in self._guides:
            dist_args = self._guides[name][1]
            for arg in dist_args:
                delattr(self, "{}_{}".format(name, arg))

        p = self._priors[name]()  # init_to_sample strategy
        if dist_constructor is dist.Delta:
            support = self._priors[name].support
            if _is_real_support(support):
                p_map = Parameter(p.detach())
            else:
                p_map = PyroParam(p.detach(), support)
            setattr(self, "{}_map".format(name), p_map)
            dist_args = ("map",)
        elif dist_constructor is dist.Normal:
            loc = Parameter(biject_to(self._priors[name].support).inv(p).detach())
            scale = PyroParam(loc.new_ones(loc.shape), constraints.positive)
            setattr(self, "{}_loc".format(name), loc)
            setattr(self, "{}_scale".format(name), scale)
            dist_args = ("loc", "scale")
        elif dist_constructor is dist.MultivariateNormal:
            loc = Parameter(biject_to(self._priors[name].support).inv(p).detach())
            identity = eye_like(loc, loc.size(-1))
            scale_tril = PyroParam(
                identity.repeat(loc.shape[:-1] + (1, 1)), constraints.lower_cholesky
            )
            setattr(self, "{}_loc".format(name), loc)
            setattr(self, "{}_scale_tril".format(name), scale_tril)
            dist_args = ("loc", "scale_tril")
        else:
            raise NotImplementedError

        self._guides[name] = (dist_constructor, dist_args)

    @pyro_method
    def _load_pyro_samples(self):
        """
        Runs `pyro.sample` primitives for all `PyroSample` attributes.
        """
        for module in self.modules():
            if "_pyro_samples" in module.__dict__:
                for name in module._pyro_samples:
                    getattr(module, name)

    def set_mode(self, mode):
        """
        Sets ``mode`` of this object to be able to use its parameters in
        stochastic functions. If ``mode="model"``, a parameter will get its
        value from its prior. If ``mode="guide"``, the value will be drawn from
        its guide.

        .. note:: This method automatically sets ``mode`` for submodules which
            belong to :class:`Parameterized` class.

        :param str mode: Either "model" or "guide".
        """
        for module in self.modules():
            if isinstance(module, Parameterized):
                module.mode = mode

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, mode):
        self._mode = mode



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import torch

import pyro
import pyro.distributions as dist
from pyro.ops.tensor_utils import safe_normalize

from .reparam import Reparam


class ProjectedNormalReparam(Reparam):
    """
    Reparametrizer for :class:`~pyro.distributions.ProjectedNormal` latent
    variables.

    This reparameterization works only for latent variables, not likelihoods.
    """

    def apply(self, msg):
        name = msg["name"]
        fn = msg["fn"]
        value = msg["value"]
        is_observed = msg["is_observed"]
        if is_observed:
            raise NotImplementedError(
                "ProjectedNormalReparam does not support observe statements"
            )

        fn, event_dim = self._unwrap(fn)
        assert isinstance(fn, dist.ProjectedNormal)

        # Differentiably invert transform.
        value_normal = None
        if value is not None:
            # We use an arbitrary injection, which works only for initialization.
            value_normal = value - fn.concentration

        # Draw parameter-free noise.
        new_fn = dist.Normal(torch.zeros_like(fn.concentration), 1).to_event(1)
        x = pyro.sample(
            "{}_normal".format(name),
            self._wrap(new_fn, event_dim),
            obs=value_normal,
            infer={"is_observed": is_observed},
        )

        # Differentiably transform.
        if value is None:
            value = safe_normalize(x + fn.concentration)

        # Simulate a pyro.deterministic() site.
        new_fn = dist.Delta(value, event_dim=event_dim).mask(False)
        return {"fn": new_fn, "value": value, "is_observed": True}



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

from pyro.distributions.coalescent import bio_phylo_to_times

from .compartmental import CompartmentalModel
from .distributions import beta_binomial_dist, binomial_dist, infection_dist

__all__ = [
    "CompartmentalModel",
    "beta_binomial_dist",
    "binomial_dist",
    "bio_phylo_to_times",
    "infection_dist",
]



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

"""
Pyro includes a class :class:`~pyro.nn.module.PyroModule`, a subclass of
:class:`torch.nn.Module`, whose attributes can be modified by Pyro effects.  To
create a poutine-aware attribute, use either the :class:`PyroParam` struct or
the :class:`PyroSample` struct::

    my_module = PyroModule()
    my_module.x = PyroParam(torch.tensor(1.), constraint=constraints.positive)
    my_module.y = PyroSample(dist.Normal(0, 1))

"""
import functools
import inspect
import weakref
from collections import OrderedDict, namedtuple

import torch
from torch.distributions import constraints, transform_to

import pyro
import pyro.params.param_store
from pyro.ops.provenance import detach_provenance
from pyro.poutine.runtime import _PYRO_PARAM_STORE

_MODULE_LOCAL_PARAMS: bool = False


@pyro.settings.register("module_local_params", __name__, "_MODULE_LOCAL_PARAMS")
def _validate_module_local_params(value: bool) -> None:
    assert isinstance(value, bool)


def _is_module_local_param_enabled() -> bool:
    return pyro.settings.get("module_local_params")


class PyroParam(namedtuple("PyroParam", ("init_value", "constraint", "event_dim"))):
    """
    Declares a Pyro-managed learnable attribute of a :class:`PyroModule`,
    similar to :func:`pyro.param <pyro.primitives.param>`.

    This can be used either to set attributes of :class:`PyroModule`
    instances::

        assert isinstance(my_module, PyroModule)
        my_module.x = PyroParam(torch.zeros(4))                   # eager
        my_module.y = PyroParam(lambda: torch.randn(4))           # lazy
        my_module.z = PyroParam(torch.ones(4),                    # eager
                                constraint=constraints.positive,
                                event_dim=1)

    or EXPERIMENTALLY as a decorator on lazy initialization properties::

        class MyModule(PyroModule):
            @PyroParam
            def x(self):
                return torch.zeros(4)

            @PyroParam
            def y(self):
                return torch.randn(4)

            @PyroParam(constraint=constraints.real, event_dim=1)
            def z(self):
                return torch.ones(4)

            def forward(self):
                return self.x + self.y + self.z  # accessed like a @property

    :param init_value: Either a tensor for eager initialization, a callable for
        lazy initialization, or None for use as a decorator.
    :type init_value: torch.Tensor or callable returning a torch.Tensor or None
    :param constraint: torch constraint, defaults to ``constraints.real``.
    :type constraint: ~torch.distributions.constraints.Constraint
    :param int event_dim: (optional) number of rightmost dimensions unrelated
        to baching. Dimension to the left of this will be considered batch
        dimensions; if the param statement is inside a subsampled plate, then
        corresponding batch dimensions of the parameter will be correspondingly
        subsampled. If unspecified, all dimensions will be considered event
        dims and no subsampling will be performed.
    """

    # Support use as a decorator.
    def __get__(self, obj, obj_type):
        assert issubclass(obj_type, PyroModule)
        if obj is None:
            return self

        name = self.init_value.__name__
        if name not in obj.__dict__["_pyro_params"]:
            init_value, constraint, event_dim = self
            init_value = functools.partial(init_value, obj)  # bind method's self arg
            setattr(obj, name, PyroParam(init_value, constraint, event_dim))
        return obj.__getattr__(name)

    # Support decoration with optional kwargs, e.g. @PyroParam(event_dim=0).
    def __call__(self, init_value):
        assert self.init_value is None
        return PyroParam(init_value, self.constraint, self.event_dim)


PyroParam.__new__.__defaults__ = (None, constraints.real, None)


class PyroSample(namedtuple("PyroSample", ("prior",))):
    """
    Declares a Pyro-managed random attribute of a :class:`PyroModule`, similar
    to :func:`pyro.sample <pyro.primitives.sample>`.

    This can be used either to set attributes of :class:`PyroModule`
    instances::

        assert isinstance(my_module, PyroModule)
        my_module.x = PyroSample(Normal(0, 1))                    # independent
        my_module.y = PyroSample(lambda self: Normal(self.x, 1))  # dependent

    or EXPERIMENTALLY as a decorator on lazy initialization methods::

        class MyModule(PyroModule):
            @PyroSample
            def x(self):
                return Normal(0, 1)       # independent

            @PyroSample
            def y(self):
                return Normal(self.x, 1)  # dependent

            def forward(self):
                return self.y             # accessed like a @property

    :param prior: distribution object or function that inputs the
        :class:`PyroModule` instance ``self`` and returns a distribution
        object.
    """

    def __init__(self, prior):
        super().__init__()
        if not hasattr(prior, "sample"):  # if not a distribution
            assert 1 == sum(
                1
                for p in inspect.signature(prior).parameters.values()
                if p.default is inspect.Parameter.empty
            ), "prior should take the single argument 'self'"
            self.name = getattr(prior, "__name__", None)
            if self.name is not None:
                # Ensure decorated function is accessible for pickling.
                prior.__name__ = "_pyro_prior_" + prior.__name__
                qualname = prior.__qualname__.rsplit(".", 1)
                qualname[-1] = prior.__name__
                prior.__qualname__ = ".".join(qualname)

    # Support use as a decorator.
    def __get__(self, obj, obj_type):
        assert issubclass(obj_type, PyroModule)
        if obj is None:
            return self

        if self.name is None:
            for name in dir(obj_type):
                if getattr(obj_type, name) is self:
                    self.name = name
                    break
        else:
            setattr(obj_type, self.prior.__name__, self.prior)  # for pickling

        obj.__dict__["_pyro_samples"].setdefault(self.name, self.prior)
        return obj.__getattr__(self.name)


def _make_name(prefix, name):
    return "{}.{}".format(prefix, name) if prefix else name


def _unconstrain(constrained_value, constraint):
    with torch.no_grad():
        if callable(constrained_value):
            constrained_value = constrained_value()
        unconstrained_value = transform_to(constraint).inv(constrained_value.detach())
        return torch.nn.Parameter(unconstrained_value)


class _Context:
    """
    Sometimes-active cache for ``PyroModule.__call__()`` contexts.
    """

    def __init__(self):
        self.active = 0
        self.cache = {}
        self.used = False
        if _is_module_local_param_enabled():
            self.param_state = {"params": {}, "constraints": {}}

    def __enter__(self):
        if not self.active and _is_module_local_param_enabled():
            self._param_ctx = pyro.get_param_store().scope(state=self.param_state)
            self.param_state = self._param_ctx.__enter__()
        self.active += 1
        self.used = True

    def __exit__(self, type, value, traceback):
        self.active -= 1
        if not self.active:
            self.cache.clear()
            if _is_module_local_param_enabled():
                self._param_ctx.__exit__(type, value, traceback)
                del self._param_ctx

    def get(self, name):
        if self.active:
            return self.cache.get(name)

    def set(self, name, value):
        if self.active:
            self.cache[name] = value


def _get_pyro_params(module):
    for name in module._parameters:
        if name.endswith("_unconstrained"):
            constrained_name = name[: -len("_unconstrained")]
            if (
                isinstance(module, PyroModule)
                and constrained_name in module._pyro_params
            ):
                yield constrained_name, getattr(module, constrained_name)
                continue
        yield name, module._parameters[name]


class _PyroModuleMeta(type):
    _pyro_mixin_cache = {}

    # Unpickling helper to create an empty object of type PyroModule[Module].
    class _New:
        def __init__(self, Module):
            self.__class__ = PyroModule[Module]

    def __getitem__(cls, Module):
        assert isinstance(Module, type)
        assert issubclass(Module, torch.nn.Module)
        if issubclass(Module, PyroModule):
            return Module
        if Module is torch.nn.Module:
            return PyroModule
        if Module in _PyroModuleMeta._pyro_mixin_cache:
            return _PyroModuleMeta._pyro_mixin_cache[Module]
        bases = [
            PyroModule[b] for b in Module.__bases__ if issubclass(b, torch.nn.Module)
        ]

        class result(Module, *bases):
            # Unpickling helper to load an object of type PyroModule[Module].
            def __reduce__(self):
                state = getattr(self, "__getstate__", self.__dict__.copy)()
                return _PyroModuleMeta._New, (Module,), state

        result.__name__ = "Pyro" + Module.__name__
        _PyroModuleMeta._pyro_mixin_cache[Module] = result
        return result


class PyroModule(torch.nn.Module, metaclass=_PyroModuleMeta):
    """
    Subclass of :class:`torch.nn.Module` whose attributes can be modified by
    Pyro effects. Attributes can be set using helpers :class:`PyroParam` and
    :class:`PyroSample` , and methods can be decorated by :func:`pyro_method` .

    **Parameters**

    To create a Pyro-managed parameter attribute, set that attribute using
    either :class:`torch.nn.Parameter` (for unconstrained parameters) or
    :class:`PyroParam` (for constrained parameters). Reading that attribute
    will then trigger a :func:`pyro.param <pyro.primitives.param>` statement.
    For example::

        # Create Pyro-managed parameter attributes.
        my_module = PyroModule()
        my_module.loc = nn.Parameter(torch.tensor(0.))
        my_module.scale = PyroParam(torch.tensor(1.),
                                    constraint=constraints.positive)
        # Read the attributes.
        loc = my_module.loc  # Triggers a pyro.param statement.
        scale = my_module.scale  # Triggers another pyro.param statement.

    Note that, unlike normal :class:`torch.nn.Module` s, :class:`PyroModule` s
    should not be registered with :func:`pyro.module <pyro.primitives.module>`
    statements.  :class:`PyroModule` s can contain other :class:`PyroModule` s
    and normal :class:`torch.nn.Module` s.  Accessing a normal
    :class:`torch.nn.Module` attribute of a :class:`PyroModule` triggers a
    :func:`pyro.module <pyro.primitives.module>` statement.  If multiple
    :class:`PyroModule` s appear in a single Pyro model or guide, they should
    be included in a single root :class:`PyroModule` for that model.

    :class:`PyroModule` s synchronize data with the param store at each
    ``setattr``, ``getattr``, and ``delattr`` event, based on the nested name
    of an attribute:

    -   Setting ``mod.x = x_init`` tries to read ``x`` from the param store. If a
        value is found in the param store, that value is copied into ``mod``
        and ``x_init`` is ignored; otherwise ``x_init`` is copied into both
        ``mod`` and the param store.
    -   Reading ``mod.x`` tries to read ``x`` from the param store. If a
        value is found in the param store, that value is copied into ``mod``;
        otherwise ``mod``'s value is copied into the param store. Finally
        ``mod`` and the param store agree on a single value to return.
    -   Deleting ``del mod.x`` removes a value from both ``mod`` and the param
        store.

    Note two :class:`PyroModule` of the same name will both synchronize with
    the global param store and thus contain the same data.  When creating a
    :class:`PyroModule`, then deleting it, then creating another with the same
    name, the latter will be populated with the former's data from the param
    store. To avoid this persistence, either ``pyro.clear_param_store()`` or
    call :func:`clear` before deleting a :class:`PyroModule` .

    :class:`PyroModule` s can be saved and loaded either directly using
    :func:`torch.save` / :func:`torch.load` or indirectly using the param
    store's :meth:`~pyro.params.param_store.ParamStoreDict.save` /
    :meth:`~pyro.params.param_store.ParamStoreDict.load` . Note that
    :func:`torch.load` will be overridden by any values in the param store, so
    it is safest to ``pyro.clear_param_store()`` before loading.

    **Samples**

    To create a Pyro-managed random attribute, set that attribute using the
    :class:`PyroSample` helper, specifying a prior distribution. Reading that
    attribute will then trigger a :func:`pyro.sample <pyro.primitives.sample>`
    statement. For example::

        # Create Pyro-managed random attributes.
        my_module.x = PyroSample(dist.Normal(0, 1))
        my_module.y = PyroSample(lambda self: dist.Normal(self.loc, self.scale))

        # Sample the attributes.
        x = my_module.x  # Triggers a pyro.sample statement.
        y = my_module.y  # Triggers one pyro.sample + two pyro.param statements.

    Sampling is cached within each invocation of ``.__call__()`` or method
    decorated by :func:`pyro_method` .  Because sample statements can appear
    only once in a Pyro trace, you should ensure that traced access to sample
    attributes is wrapped in a single invocation of ``.__call__()`` or method
    decorated by :func:`pyro_method` .

    To make an existing module probabilistic, you can create a subclass and
    overwrite some parameters with :class:`PyroSample` s::

        class RandomLinear(nn.Linear, PyroModule):  # used as a mixin
            def __init__(self, in_features, out_features):
                super().__init__(in_features, out_features)
                self.weight = PyroSample(
                    lambda self: dist.Normal(0, 1)
                                     .expand([self.out_features,
                                              self.in_features])
                                     .to_event(2))

    **Mixin classes**

    :class:`PyroModule` can be used as a mixin class, and supports simple
    syntax for dynamically creating mixins, for example the following are
    equivalent::

        # Version 1. create a named mixin class
        class PyroLinear(nn.Linear, PyroModule):
            pass

        m.linear = PyroLinear(m, n)

        # Version 2. create a dynamic mixin class
        m.linear = PyroModule[nn.Linear](m, n)

    This notation can be used recursively to create Bayesian modules, e.g.::

        model = PyroModule[nn.Sequential](
            PyroModule[nn.Linear](28 * 28, 100),
            PyroModule[nn.Sigmoid](),
            PyroModule[nn.Linear](100, 100),
            PyroModule[nn.Sigmoid](),
            PyroModule[nn.Linear](100, 10),
        )
        assert isinstance(model, nn.Sequential)
        assert isinstance(model, PyroModule)

        # Now we can be Bayesian about weights in the first layer.
        model[0].weight = PyroSample(
            prior=dist.Normal(0, 1).expand([28 * 28, 100]).to_event(2))
        guide = AutoDiagonalNormal(model)

    Note that ``PyroModule[...]`` does not recursively mix in
    :class:`PyroModule` to submodules of the input ``Module``; hence we needed
    to wrap each submodule of the ``nn.Sequential`` above.

    :param str name: Optional name for a root PyroModule. This is ignored in
        sub-PyroModules of another PyroModule.
    """

    def __init__(self, name=""):
        self._pyro_name = name
        self._pyro_context = _Context()  # shared among sub-PyroModules
        self._pyro_params = OrderedDict()
        self._pyro_samples = OrderedDict()
        super().__init__()

    def add_module(self, name, module):
        """
        Adds a child module to the current module.
        """
        if isinstance(module, PyroModule):
            module._pyro_set_supermodule(
                _make_name(self._pyro_name, name), self._pyro_context
            )
        super().add_module(name, module)

    def named_pyro_params(self, prefix="", recurse=True):
        """
        Returns an iterator over PyroModule parameters, yielding both the
        name of the parameter as well as the parameter itself.

        :param str prefix: prefix to prepend to all parameter names.
        :param bool recurse: if True, then yields parameters of this module
            and all submodules. Otherwise, yields only parameters that
            are direct members of this module.
        :returns: a generator which yields tuples containing the name and parameter
        """
        gen = self._named_members(_get_pyro_params, prefix=prefix, recurse=recurse)
        for elem in gen:
            yield elem

    def _pyro_set_supermodule(self, name, context):
        if _is_module_local_param_enabled() and pyro.settings.get("validate_poutine"):
            self._check_module_local_param_usage()
        self._pyro_name = name
        self._pyro_context = context
        for key, value in self._modules.items():
            if isinstance(value, PyroModule):
                assert (
                    not value._pyro_context.used
                ), "submodule {} has executed outside of supermodule".format(name)
                value._pyro_set_supermodule(_make_name(name, key), context)

    def _pyro_get_fullname(self, name):
        assert self.__dict__["_pyro_context"].used, "fullname is not yet defined"
        return _make_name(self.__dict__["_pyro_name"], name)

    def __call__(self, *args, **kwargs):
        with self._pyro_context:
            result = super().__call__(*args, **kwargs)
        if (
            pyro.settings.get("validate_poutine")
            and not self._pyro_context.active
            and _is_module_local_param_enabled()
        ):
            self._check_module_local_param_usage()
        return result

    def _check_module_local_param_usage(self) -> None:
        self_nn_params = set(id(p) for p in self.parameters())
        self_pyro_params = set(
            id(p if not hasattr(p, "unconstrained") else p.unconstrained())
            for p in self._pyro_context.param_state["params"].values()
        )
        if not self_pyro_params <= self_nn_params:
            raise NotImplementedError(
                "Support for global pyro.param statements in PyroModules "
                "with local param mode enabled is not yet implemented."
            )

    def __getattr__(self, name):
        # PyroParams trigger pyro.param statements.
        if "_pyro_params" in self.__dict__:
            _pyro_params = self.__dict__["_pyro_params"]
            if name in _pyro_params:
                constraint, event_dim = _pyro_params[name]
                unconstrained_value = getattr(self, name + "_unconstrained")
                if self._pyro_context.active and not _is_module_local_param_enabled():
                    fullname = self._pyro_get_fullname(name)
                    if fullname in _PYRO_PARAM_STORE:
                        if (
                            _PYRO_PARAM_STORE._params[fullname]
                            is not unconstrained_value
                        ):
                            # Update PyroModule <--- ParamStore.
                            unconstrained_value = _PYRO_PARAM_STORE._params[fullname]
                            if not isinstance(unconstrained_value, torch.nn.Parameter):
                                # Update PyroModule ---> ParamStore (type only; data is preserved).
                                unconstrained_value = torch.nn.Parameter(
                                    unconstrained_value
                                )
                                _PYRO_PARAM_STORE._params[fullname] = (
                                    unconstrained_value
                                )
                                _PYRO_PARAM_STORE._param_to_name[
                                    unconstrained_value
                                ] = fullname
                            super().__setattr__(
                                name + "_unconstrained", unconstrained_value
                            )
                    else:
                        # Update PyroModule ---> ParamStore.
                        _PYRO_PARAM_STORE._constraints[fullname] = constraint
                        _PYRO_PARAM_STORE._params[fullname] = unconstrained_value
                        _PYRO_PARAM_STORE._param_to_name[unconstrained_value] = fullname
                    return pyro.param(fullname, event_dim=event_dim)
                elif self._pyro_context.active and _is_module_local_param_enabled():
                    # fake param statement to ensure any handlers of pyro.param are applied,
                    # even though we don't use the contents of the local parameter store
                    fullname = self._pyro_get_fullname(name)
                    constrained_value = transform_to(constraint)(unconstrained_value)
                    constrained_value.unconstrained = weakref.ref(unconstrained_value)
                    return pyro.poutine.runtime.effectful(type="param")(
                        lambda *_, **__: constrained_value
                    )(fullname, event_dim=event_dim, name=fullname)
                else:  # Cannot determine supermodule and hence cannot compute fullname.
                    constrained_value = transform_to(constraint)(unconstrained_value)
                    constrained_value.unconstrained = weakref.ref(unconstrained_value)
                    return constrained_value

        # PyroSample trigger pyro.sample statements.
        if "_pyro_samples" in self.__dict__:
            _pyro_samples = self.__dict__["_pyro_samples"]
            if name in _pyro_samples:
                prior = _pyro_samples[name]
                context = self._pyro_context
                if context.active:
                    fullname = self._pyro_get_fullname(name)
                    value = context.get(fullname)
                    if value is None:
                        if not hasattr(prior, "sample"):  # if not a distribution
                            prior = prior(self)
                        value = pyro.sample(fullname, prior)
                        context.set(fullname, value)
                    return value
                else:  # Cannot determine supermodule and hence cannot compute fullname.
                    if not hasattr(prior, "sample"):  # if not a distribution
                        prior = prior(self)
                    return prior()

        result = super().__getattr__(name)

        # Regular nn.Parameters trigger pyro.param statements.
        if isinstance(result, torch.nn.Parameter) and not name.endswith(
            "_unconstrained"
        ):
            if self._pyro_context.active and not _is_module_local_param_enabled():
                pyro.param(self._pyro_get_fullname(name), result)
            elif self._pyro_context.active and _is_module_local_param_enabled():
                # fake param statement to ensure any handlers of pyro.param are applied,
                # even though we don't use the contents of the local parameter store
                fullname = self._pyro_get_fullname(name)
                pyro.poutine.runtime.effectful(type="param")(lambda *_, **__: result)(
                    fullname, result, name=fullname
                )

        if isinstance(result, torch.nn.Module):
            if isinstance(result, PyroModule):
                if not result._pyro_name:
                    # Update sub-PyroModules that were converted from nn.Modules in-place.
                    result._pyro_set_supermodule(
                        _make_name(self._pyro_name, name), self._pyro_context
                    )
            else:
                # Regular nn.Modules trigger pyro.module statements.
                if self._pyro_context.active and not _is_module_local_param_enabled():
                    pyro.module(self._pyro_get_fullname(name), result)
                elif self._pyro_context.active and _is_module_local_param_enabled():
                    # fake module statement to ensure any handlers of pyro.module are applied,
                    # even though we don't use the contents of the local parameter store
                    fullname_module = self._pyro_get_fullname(name)
                    for param_name, param_value in result.named_parameters():
                        fullname_param = pyro.params.param_store.param_with_module_name(
                            fullname_module, param_name
                        )
                        pyro.poutine.runtime.effectful(type="param")(
                            lambda *_, **__: param_value
                        )(fullname_param, param_value, name=fullname_param)

        return result

    def __setattr__(self, name, value):
        if isinstance(value, PyroModule):
            # Create a new sub PyroModule, overwriting any old value.
            try:
                delattr(self, name)
            except AttributeError:
                pass
            self.add_module(name, value)
            return

        if isinstance(value, PyroParam):
            # Create a new PyroParam, overwriting any old value.
            try:
                delattr(self, name)
            except AttributeError:
                pass
            constrained_value, constraint, event_dim = value
            self._pyro_params[name] = constraint, event_dim
            if self._pyro_context.active and not _is_module_local_param_enabled():
                fullname = self._pyro_get_fullname(name)
                pyro.param(
                    fullname,
                    constrained_value,
                    constraint=constraint,
                    event_dim=event_dim,
                )
                constrained_value = detach_provenance(pyro.param(fullname))
                unconstrained_value = constrained_value.unconstrained()
                if not isinstance(unconstrained_value, torch.nn.Parameter):
                    # Update PyroModule ---> ParamStore (type only; data is preserved).
                    unconstrained_value = torch.nn.Parameter(unconstrained_value)
                    _PYRO_PARAM_STORE._params[fullname] = unconstrained_value
                    _PYRO_PARAM_STORE._param_to_name[unconstrained_value] = fullname
            elif self._pyro_context.active and _is_module_local_param_enabled():
                # fake param statement to ensure any handlers of pyro.param are applied,
                # even though we don't use the contents of the local parameter store
                fullname = self._pyro_get_fullname(name)
                constrained_value = detach_provenance(
                    pyro.poutine.runtime.effectful(type="param")(
                        lambda *_, **__: constrained_value
                    )(
                        fullname,
                        constraint=constraint,
                        event_dim=event_dim,
                        name=fullname,
                    )
                )
                unconstrained_value = _unconstrain(constrained_value, constraint)
            else:  # Cannot determine supermodule and hence cannot compute fullname.
                unconstrained_value = _unconstrain(constrained_value, constraint)
            super().__setattr__(name + "_unconstrained", unconstrained_value)
            return

        if isinstance(value, torch.nn.Parameter):
            # Create a new nn.Parameter, overwriting any old value.
            try:
                delattr(self, name)
            except AttributeError:
                pass
            if self._pyro_context.active and not _is_module_local_param_enabled():
                fullname = self._pyro_get_fullname(name)
                value = pyro.param(fullname, value)
                if not isinstance(value, torch.nn.Parameter):
                    # Update PyroModule ---> ParamStore (type only; data is preserved).
                    value = torch.nn.Parameter(detach_provenance(value))
                    _PYRO_PARAM_STORE._params[fullname] = value
                    _PYRO_PARAM_STORE._param_to_name[value] = fullname
            elif self._pyro_context.active and _is_module_local_param_enabled():
                # fake param statement to ensure any handlers of pyro.param are applied,
                # even though we don't use the contents of the local parameter store
                fullname = self._pyro_get_fullname(name)
                value = detach_provenance(
                    pyro.poutine.runtime.effectful(type="param")(
                        lambda *_, **__: value
                    )(fullname, value, constraint=constraints.real, name=fullname)
                )
            super().__setattr__(name, value)
            return

        if isinstance(value, torch.Tensor):
            if name in self._pyro_params:
                # Update value of an existing PyroParam.
                constraint, event_dim = self._pyro_params[name]
                unconstrained_value = getattr(self, name + "_unconstrained")
                with torch.no_grad():
                    unconstrained_value.data = transform_to(constraint).inv(
                        value.detach()
                    )
                return

        if isinstance(value, PyroSample):
            # Create a new PyroSample, overwriting any old value.
            try:
                delattr(self, name)
            except AttributeError:
                pass
            _pyro_samples = self.__dict__["_pyro_samples"]
            _pyro_samples[name] = value.prior
            return

        super().__setattr__(name, value)

    def __delattr__(self, name):
        if name in self._parameters:
            del self._parameters[name]
            if self._pyro_context.used:
                fullname = self._pyro_get_fullname(name)
                if fullname in _PYRO_PARAM_STORE:
                    # Update PyroModule ---> ParamStore.
                    del _PYRO_PARAM_STORE[fullname]
            return

        if name in self._pyro_params:
            delattr(self, name + "_unconstrained")
            del self._pyro_params[name]
            if self._pyro_context.used:
                fullname = self._pyro_get_fullname(name)
                if fullname in _PYRO_PARAM_STORE:
                    # Update PyroModule ---> ParamStore.
                    del _PYRO_PARAM_STORE[fullname]
            return

        if name in self._pyro_samples:
            del self._pyro_samples[name]
            return

        if name in self._modules:
            del self._modules[name]
            if self._pyro_context.used:
                fullname = self._pyro_get_fullname(name)
                for p in list(_PYRO_PARAM_STORE.keys()):
                    if p.startswith(fullname):
                        del _PYRO_PARAM_STORE[p]
            return

        super().__delattr__(name)

    def __getstate__(self):
        # Remove weakrefs in preparation for pickling.
        for param in self.parameters(recurse=True):
            param.__dict__.pop("unconstrained", None)
        return getattr(super(), "__getstate__", self.__dict__.copy)()


def pyro_method(fn):
    """
    Decorator for top-level methods of a :class:`PyroModule` to enable pyro
    effects and cache ``pyro.sample`` statements.

    This should be applied to all public methods that read Pyro-managed
    attributes, but is not needed for ``.forward()``.
    """

    @functools.wraps(fn)
    def cached_fn(self, *args, **kwargs):
        with self._pyro_context:
            return fn(self, *args, **kwargs)

    return cached_fn


def clear(mod):
    """
    Removes data from both a :class:`PyroModule` and the param store.

    :param PyroModule mod: A module to clear.
    """
    assert isinstance(mod, PyroModule)
    for name in list(mod._pyro_params):
        delattr(mod, name)
    for name in list(mod._parameters):
        delattr(mod, name)
    for name in list(mod._modules):
        delattr(mod, name)


def to_pyro_module_(m, recurse=True):
    """
    Converts an ordinary :class:`torch.nn.Module` instance to a
    :class:`PyroModule` **in-place**.

    This is useful for adding Pyro effects to third-party modules: no
    third-party code needs to be modified. For example::

        model = nn.Sequential(
            nn.Linear(28 * 28, 100),
            nn.Sigmoid(),
            nn.Linear(100, 100),
            nn.Sigmoid(),
            nn.Linear(100, 10),
        )
        to_pyro_module_(model)
        assert isinstance(model, PyroModule[nn.Sequential])
        assert isinstance(model[0], PyroModule[nn.Linear])

        # Now we can attempt to be fully Bayesian:
        for m in model.modules():
            for name, value in list(m.named_parameters(recurse=False)):
                setattr(m, name, PyroSample(prior=dist.Normal(0, 1)
                                                      .expand(value.shape)
                                                      .to_event(value.dim())))
        guide = AutoDiagonalNormal(model)

    :param torch.nn.Module m: A module instance.
    :param bool recurse: Whether to convert submodules to :class:`PyroModules` .
    """
    if not isinstance(m, torch.nn.Module):
        raise TypeError("Expected an nn.Module instance but got a {}".format(type(m)))

    if isinstance(m, PyroModule):
        if recurse:
            for name, value in list(m._modules.items()):
                to_pyro_module_(value)
                setattr(m, name, value)
        return

    # Change m's type in-place.
    m.__class__ = PyroModule[m.__class__]
    m._pyro_name = ""
    m._pyro_context = _Context()
    m._pyro_params = OrderedDict()
    m._pyro_samples = OrderedDict()

    # Reregister parameters and submodules.
    for name, value in list(m._parameters.items()):
        setattr(m, name, value)
    for name, value in list(m._modules.items()):
        if recurse:
            to_pyro_module_(value)
        setattr(m, name, value)


# The following descriptor disables the ._flat_weights cache of
# torch.nn.RNNBase, forcing recomputation on each access of the ._flat_weights
# attribute. This is required if any attribute is set to a PyroParam or
# PyroSample. For motivation, see https://github.com/pyro-ppl/pyro/issues/2390
class _FlatWeightsDescriptor:
    def __get__(self, obj, obj_type=None):
        if obj is None:
            return self
        return [getattr(obj, name) for name in obj._flat_weights_names]

    def __set__(self, obj, value):
        pass  # Ignore value.


PyroModule[torch.nn.RNNBase]._flat_weights = _FlatWeightsDescriptor()



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

from functools import partial

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions.transforms import SigmoidTransform, TanhTransform

from pyro.nn import AutoRegressiveNN, ConditionalAutoRegressiveNN

from .. import constraints
from ..conditional import ConditionalTransformModule
from ..torch_transform import TransformModule
from ..util import copy_docs_from
from .basic import ELUTransform, LeakyReLUTransform

eps = 1e-8


@copy_docs_from(TransformModule)
class NeuralAutoregressive(TransformModule):
    r"""
    An implementation of the deep Neural Autoregressive Flow (NAF) bijective
    transform of the "IAF flavour" that can be used for sampling and scoring samples
    drawn from it (but not arbitrary ones).

    Example usage:

    >>> from pyro.nn import AutoRegressiveNN
    >>> base_dist = dist.Normal(torch.zeros(10), torch.ones(10))
    >>> arn = AutoRegressiveNN(10, [40], param_dims=[16]*3)
    >>> transform = NeuralAutoregressive(arn, hidden_units=16)
    >>> pyro.module("my_transform", transform)  # doctest: +SKIP
    >>> flow_dist = dist.TransformedDistribution(base_dist, [transform])
    >>> flow_dist.sample()  # doctest: +SKIP

    The inverse operation is not implemented. This would require numerical
    inversion, e.g., using a root finding method - a possibility for a future
    implementation.

    :param autoregressive_nn: an autoregressive neural network whose forward call
        returns a tuple of three real-valued tensors, whose last dimension is the
        input dimension, and whose penultimate dimension is equal to hidden_units.
    :type autoregressive_nn: nn.Module
    :param hidden_units: the number of hidden units to use in the NAF transformation
        (see Eq (8) in reference)
    :type hidden_units: int
    :param activation: Activation function to use. One of 'ELU', 'LeakyReLU',
        'sigmoid', or 'tanh'.
    :type activation: string

    Reference:

    [1] Chin-Wei Huang, David Krueger, Alexandre Lacoste, Aaron Courville. Neural
    Autoregressive Flows. [arXiv:1804.00779]


    """

    domain = constraints.real_vector
    codomain = constraints.real_vector
    bijective = True
    eps = 1e-8
    autoregressive = True

    def __init__(self, autoregressive_nn, hidden_units=16, activation="sigmoid"):
        super().__init__(cache_size=1)

        # Create the intermediate transform used
        name_to_mixin = {
            "ELU": ELUTransform,
            "LeakyReLU": LeakyReLUTransform,
            "sigmoid": SigmoidTransform,
            "tanh": TanhTransform,
        }
        if activation not in name_to_mixin:
            raise ValueError('Invalid activation function "{}"'.format(activation))
        self.T = name_to_mixin[activation]()

        self.arn = autoregressive_nn
        self.hidden_units = hidden_units
        self.logsoftmax = nn.LogSoftmax(dim=-2)
        self._cached_log_df_inv_dx = None
        self._cached_A = None
        self._cached_W_pre = None
        self._cached_C = None
        self._cached_T_C = None

    def _call(self, x):
        """
        :param x: the input into the bijection
        :type x: torch.Tensor

        Invokes the bijection x=>y; in the prototypical context of a
        :class:`~pyro.distributions.TransformedDistribution` `x` is a sample from
        the base distribution (or the output of a previous transform)
        """
        # A, W, b ~ batch_shape x hidden_units x event_shape
        A, W_pre, b = self.arn(x)
        T = self.T

        # Divide the autoregressive output into the component activations
        A = F.softplus(A)
        C = A * x.unsqueeze(-2) + b
        W = F.softmax(W_pre, dim=-2)
        T_C = T(C)
        D = (W * T_C).sum(dim=-2)
        y = T.inv(D)

        self._cached_log_df_inv_dx = T.inv.log_abs_det_jacobian(D, y)
        self._cached_A = A
        self._cached_W_pre = W_pre
        self._cached_C = C
        self._cached_T_C = T_C

        return y

    # This method returns log(abs(det(dy/dx)), which is equal to -log(abs(det(dx/dy))
    def log_abs_det_jacobian(self, x, y):
        """
        Calculates the elementwise determinant of the log Jacobian
        """

        A = self._cached_A
        W_pre = self._cached_W_pre
        C = self._cached_C
        T_C = self._cached_T_C
        T = self.T

        log_dydD = self._cached_log_df_inv_dx
        log_dDdx = torch.logsumexp(
            torch.log(A + self.eps)
            + self.logsoftmax(W_pre)
            + T.log_abs_det_jacobian(C, T_C),
            dim=-2,
        )
        log_det = log_dydD + log_dDdx
        return log_det.sum(-1)


@copy_docs_from(ConditionalTransformModule)
class ConditionalNeuralAutoregressive(ConditionalTransformModule):
    r"""
    An implementation of the deep Neural Autoregressive Flow (NAF) bijective
    transform of the "IAF flavour" conditioning on an additiona context variable
    that can be used for sampling and scoring samples drawn from it (but not
    arbitrary ones).

    Example usage:

    >>> from pyro.nn import ConditionalAutoRegressiveNN
    >>> input_dim = 10
    >>> context_dim = 5
    >>> batch_size = 3
    >>> base_dist = dist.Normal(torch.zeros(input_dim), torch.ones(input_dim))
    >>> arn = ConditionalAutoRegressiveNN(input_dim, context_dim, [40],
    ... param_dims=[16]*3)
    >>> transform = ConditionalNeuralAutoregressive(arn, hidden_units=16)
    >>> pyro.module("my_transform", transform)  # doctest: +SKIP
    >>> z = torch.rand(batch_size, context_dim)
    >>> flow_dist = dist.ConditionalTransformedDistribution(base_dist,
    ... [transform]).condition(z)
    >>> flow_dist.sample(sample_shape=torch.Size([batch_size]))  # doctest: +SKIP

    The inverse operation is not implemented. This would require numerical
    inversion, e.g., using a root finding method - a possibility for a future
    implementation.

    :param autoregressive_nn: an autoregressive neural network whose forward call
        returns a tuple of three real-valued tensors, whose last dimension is the
        input dimension, and whose penultimate dimension is equal to hidden_units.
    :type autoregressive_nn: nn.Module
    :param hidden_units: the number of hidden units to use in the NAF transformation
        (see Eq (8) in reference)
    :type hidden_units: int
    :param activation: Activation function to use. One of 'ELU', 'LeakyReLU',
        'sigmoid', or 'tanh'.
    :type activation: string

    Reference:

    [1] Chin-Wei Huang, David Krueger, Alexandre Lacoste, Aaron Courville. Neural
    Autoregressive Flows. [arXiv:1804.00779]


    """

    domain = constraints.real_vector
    codomain = constraints.real_vector
    bijective = True

    def __init__(self, autoregressive_nn, **kwargs):
        super().__init__()
        self.nn = autoregressive_nn
        self.kwargs = kwargs

    def condition(self, context):
        """
        Conditions on a context variable, returning a non-conditional transform of
        of type :class:`~pyro.distributions.transforms.NeuralAutoregressive`.
        """

        # Note that nn.condition doesn't copy the weights of the ConditionalAutoregressiveNN
        cond_nn = partial(self.nn, context=context)
        cond_nn.permutation = cond_nn.func.permutation
        cond_nn.get_permutation = cond_nn.func.get_permutation
        return NeuralAutoregressive(cond_nn, **self.kwargs)


def neural_autoregressive(input_dim, hidden_dims=None, activation="sigmoid", width=16):
    """
    A helper function to create a
    :class:`~pyro.distributions.transforms.NeuralAutoregressive` object that takes
    care of constructing an autoregressive network with the correct input/output
    dimensions.

    :param input_dim: Dimension of input variable
    :type input_dim: int
    :param hidden_dims: The desired hidden dimensions of the autoregressive network.
        Defaults to using [3*input_dim + 1]
    :type hidden_dims: list[int]
    :param activation: Activation function to use. One of 'ELU', 'LeakyReLU',
        'sigmoid', or 'tanh'.
    :type activation: string
    :param width: The width of the "multilayer perceptron" in the transform (see
        paper). Defaults to 16
    :type width: int

    """

    if hidden_dims is None:
        hidden_dims = [3 * input_dim + 1]
    arn = AutoRegressiveNN(input_dim, hidden_dims, param_dims=[width] * 3)
    return NeuralAutoregressive(arn, hidden_units=width, activation=activation)


def conditional_neural_autoregressive(
    input_dim, context_dim, hidden_dims=None, activation="sigmoid", width=16
):
    """
    A helper function to create a
    :class:`~pyro.distributions.transforms.ConditionalNeuralAutoregressive` object
    that takes care of constructing an autoregressive network with the correct
    input/output dimensions.

    :param input_dim: Dimension of input variable
    :type input_dim: int
    :param context_dim: Dimension of context variable
    :type context_dim: int
    :param hidden_dims: The desired hidden dimensions of the autoregressive network.
        Defaults to using [3*input_dim + 1]
    :type hidden_dims: list[int]
    :param activation: Activation function to use. One of 'ELU', 'LeakyReLU',
        'sigmoid', or 'tanh'.
    :type activation: string
    :param width: The width of the "multilayer perceptron" in the transform (see
        paper). Defaults to 16
    :type width: int

    """

    if hidden_dims is None:
        hidden_dims = [3 * input_dim + 1]
    arn = ConditionalAutoRegressiveNN(
        input_dim, context_dim, hidden_dims, param_dims=[width] * 3
    )
    return ConditionalNeuralAutoregressive(
        arn, hidden_units=width, activation=activation
    )



# Copyright Contributors to the Pyro project.
# Copyright (c) 2020, YosefLab.
# SPDX-License-Identifier: Apache-2.0 AND BSD-3-Clause

"""
The data preprocessing code in this script is adapted from:
https://github.com/YosefLab/scvi-tutorials/blob/50dd3269abfe0c375ec47114f2c20725a016736f/seed_labeling.ipynb
"""

import math

import numpy as np
import torch
import torch.nn as nn
from scipy import sparse


class BatchDataLoader(object):
    """
    This custom DataLoader serves mini-batches that are either fully-observed (i.e. labeled)
    or partially-observed (i.e. unlabeled) but never mixed.
    """

    def __init__(self, data_x, data_y, batch_size, num_classes=4, missing_label=-1):
        super().__init__()
        self.data_x = data_x
        self.data_y = data_y
        self.batch_size = batch_size
        self.num_classes = num_classes

        self.unlabeled = torch.where(data_y == missing_label)[0]
        self.num_unlabeled = self.unlabeled.size(0)
        self.num_unlabeled_batches = math.ceil(self.num_unlabeled / self.batch_size)

        self.labeled = torch.where(data_y != missing_label)[0]
        self.num_labeled = self.labeled.size(0)
        self.num_labeled_batches = math.ceil(self.num_labeled / self.batch_size)

        assert self.data_x.size(0) == self.data_y.size(0)
        assert len(self) > 0

    @property
    def size(self):
        return self.data_x.size(0)

    def __len__(self):
        return self.num_unlabeled_batches + self.num_labeled_batches

    def _sample_batch_indices(self):
        batch_order = torch.randperm(len(self)).tolist()
        unlabeled_idx = self.unlabeled[torch.randperm(self.num_unlabeled)]
        labeled_idx = self.labeled[torch.randperm(self.num_labeled)]

        slices = []

        for i in range(self.num_unlabeled_batches):
            _slice = unlabeled_idx[i * self.batch_size : (i + 1) * self.batch_size]
            slices.append((_slice, False))

        for i in range(self.num_labeled_batches):
            _slice = labeled_idx[i * self.batch_size : (i + 1) * self.batch_size]
            slices.append((_slice, True))

        return slices, batch_order

    def __iter__(self):
        slices, batch_order = self._sample_batch_indices()

        for i in range(len(batch_order)):
            _slice = slices[batch_order[i]]
            if _slice[1]:
                # labeled
                yield self.data_x[_slice[0]], nn.functional.one_hot(
                    self.data_y[_slice[0]], num_classes=self.num_classes
                )
            else:
                # unlabeled
                yield self.data_x[_slice[0]], None


def _get_score(normalized_adata, gene_set):
    """
    Returns the score per cell given a dictionary of + and - genes
    """
    score = np.zeros(normalized_adata.n_obs)
    for gene in gene_set["positive"]:
        expression = np.array(normalized_adata[:, gene].X)
        score += expression.flatten()
    for gene in gene_set["negative"]:
        expression = np.array(normalized_adata[:, gene].X)
        score -= expression.flatten()
    return score


def _get_cell_mask(normalized_adata, gene_set):
    """
    Calculates the score per cell for a list of genes, then returns a mask for
    the cells with the highest 50 scores.
    """
    score = _get_score(normalized_adata, gene_set)
    cell_idx = score.argsort()[-50:]
    mask = np.zeros(normalized_adata.n_obs)
    mask[cell_idx] = 1
    return mask.astype(bool)


def get_data(dataset="pbmc", batch_size=100, cuda=False):
    """
    Does the necessary preprocessing and returns a BatchDataLoader for the PBMC dataset.
    """
    assert dataset in ["pbmc", "mock"]

    # create mock dataset for CI
    if dataset == "mock":
        num_genes = 17
        num_data = 200
        X = torch.distributions.Poisson(rate=10.0).sample(
            sample_shape=(num_data, num_genes)
        )
        Y = torch.zeros(num_data, dtype=torch.long)
        Y[50:100] = 1
        Y[100:] = -1

        if cuda:
            X, Y = X.cuda(), Y.cuda()

        return BatchDataLoader(X, Y, batch_size), num_genes, 2.0, 1.0, None

    import scanpy as sc
    import scvi

    adata = scvi.data.purified_pbmc_dataset(
        subset_datasets=["regulatory_t", "naive_t", "memory_t", "naive_cytotoxic"]
    )

    gene_subset = [
        "CD4",
        "FOXP3",
        "TNFRSF18",
        "IL2RA",
        "CTLA4",
        "CD44",
        "TCF7",
        "CD8B",
        "CCR7",
        "CD69",
        "PTPRC",
        "S100A4",
    ]

    normalized = adata.copy()
    sc.pp.normalize_total(normalized, target_sum=1e4)
    sc.pp.log1p(normalized)

    normalized = normalized[:, gene_subset].copy()
    sc.pp.scale(normalized)

    # hand curated list of genes for identifying ground truth
    cd4_reg_geneset = {
        "positive": ["TNFRSF18", "CTLA4", "FOXP3", "IL2RA"],
        "negative": ["S100A4", "PTPRC", "CD8B"],
    }
    cd8_naive_geneset = {"positive": ["CD8B", "CCR7"], "negative": ["CD4"]}
    cd4_naive_geneset = {
        "positive": ["CCR7", "CD4"],
        "negative": ["S100A4", "PTPRC", "FOXP3", "IL2RA", "CD69"],
    }
    cd4_mem_geneset = {
        "positive": ["S100A4"],
        "negative": ["IL2RA", "FOXP3", "TNFRSF18", "CCR7"],
    }

    cd4_reg_mask = _get_cell_mask(normalized, cd4_reg_geneset)
    cd8_naive_mask = _get_cell_mask(normalized, cd8_naive_geneset)
    cd4_naive_mask = _get_cell_mask(normalized, cd4_naive_geneset)
    cd4_mem_mask = _get_cell_mask(normalized, cd4_mem_geneset)

    # these will be our seed labels
    seed_labels = -np.ones(cd4_mem_mask.shape[0])
    seed_labels[cd8_naive_mask] = 0  # "CD8 Naive T cell"
    seed_labels[cd4_naive_mask] = 1  # "CD4 Naive T cell"
    seed_labels[cd4_mem_mask] = 2  # "CD4 Memory T cell"
    seed_labels[cd4_reg_mask] = 3  # "CD4 Regulatory T cell"

    # this metadata will be used for plotting
    seed_colors = ["lightgray"] * seed_labels.shape[0]
    seed_sizes = [0.05] * seed_labels.shape[0]
    for i in range(len(seed_colors)):
        if seed_labels[i] == 0:
            seed_colors[i] = "lightcoral"
        elif seed_labels[i] == 1:
            seed_colors[i] = "limegreen"
        elif seed_labels[i] == 2:
            seed_colors[i] = "deepskyblue"
        elif seed_labels[i] == 3:
            seed_colors[i] = "mediumorchid"
        if seed_labels[i] != -1:
            seed_sizes[i] = 25

    adata.obs["seed_labels"] = seed_labels
    adata.obs["seed_colors"] = seed_colors
    adata.obs["seed_marker_sizes"] = seed_sizes

    Y = torch.from_numpy(seed_labels).long()
    X = torch.from_numpy(sparse.csr_matrix.todense(adata.X)).float()

    # the prior mean and scale for the log count latent variable `l`
    # is set using the empirical mean and variance of the observed log counts
    log_counts = X.sum(-1).log()
    l_mean, l_scale = log_counts.mean().item(), log_counts.std().item()

    if cuda:
        X, Y = X.cuda(), Y.cuda()

    # subsample and remove ~50% of the unlabeled cells
    torch.manual_seed(0)
    labeled = torch.where(Y != -1)[0]
    unlabeled = torch.where(Y == -1)[0]
    unlabeled = unlabeled[torch.randperm(unlabeled.size(0))[:19800]]
    idx = torch.cat([labeled, unlabeled])

    num_genes = X.size(-1)

    adata = adata[idx.data.cpu().numpy()]
    adata.raw = adata

    return (
        BatchDataLoader(X[idx], Y[idx], batch_size),
        num_genes,
        l_mean,
        l_scale,
        adata,
    )



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import warnings

from pyro.infer.autoguide import *  # noqa F403

warnings.warn(
    "pyro.contrib.autoguide has moved to pyro.infer.autoguide. "
    "The contrib alias will stop working in Pyro 0.5.",
    DeprecationWarning,
)



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import pyroapi

from pyro.contrib.funsor.handlers import condition, do, markov, vectorized_markov
from pyro.contrib.funsor.handlers import plate as _plate
from pyro.contrib.funsor.handlers.primitives import to_data, to_funsor
from pyro.primitives import (
    clear_param_store,
    deterministic,
    enable_validation,
    factor,
    get_param_store,
    module,
    param,
    random_module,
    sample,
    set_rng_seed,
    subsample,
)


def plate(*args, **kwargs):
    return _plate(None, *args, **kwargs)


pyroapi.register_backend(
    "contrib.funsor",
    {
        "distributions": "pyro.distributions",
        "handlers": "pyro.contrib.funsor.handlers",
        "infer": "pyro.contrib.funsor.infer",
        "ops": "torch",
        "optim": "pyro.optim",
        "pyro": "pyro.contrib.funsor",
    },
)

__all__ = [
    "clear_param_store",
    "condition",
    "deterministic",
    "do",
    "enable_validation",
    "factor",
    "get_param_store",
    "markov",
    "module",
    "param",
    "random_module",
    "sample",
    "set_rng_seed",
    "subsample",
    "to_data",
    "to_funsor",
    "vectorized_markov",
]



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

from abc import ABCMeta, abstractmethod

import torch
from torch import nn
from torch.nn import Parameter

import pyro.distributions as dist
from pyro.distributions.util import eye_like


class DynamicModel(nn.Module, metaclass=ABCMeta):
    """
    Dynamic model interface.

    :param dimension: native state dimension.
    :param dimension_pv: PV state dimension.
    :param num_process_noise_parameters: process noise parameter space dimension.
          This for UKF applications. Can be left as ``None`` for EKF and most
          other filters.
    """

    def __init__(self, dimension, dimension_pv, num_process_noise_parameters=None):
        self._dimension = dimension
        self._dimension_pv = dimension_pv
        self._num_process_noise_parameters = num_process_noise_parameters
        super().__init__()

    @property
    def dimension(self):
        """
        Native state dimension access.
        """
        return self._dimension

    @property
    def dimension_pv(self):
        """
        PV state dimension access.
        """
        return self._dimension_pv

    @property
    def num_process_noise_parameters(self):
        """
        Process noise parameters space dimension access.
        """
        return self._num_process_noise_parameters

    @abstractmethod
    def forward(self, x, dt, do_normalization=True):
        """
        Integrate native state ``x`` over time interval ``dt``.

        :param x: current native state. If the DynamicModel is non-differentiable,
              be sure to handle the case of ``x`` being augmented with process
              noise parameters.
        :param dt: time interval to integrate over.
        :param do_normalization: whether to perform normalization on output, e.g.,
              mod'ing angles into an interval.
        :return: Native state x integrated dt into the future.
        """
        raise NotImplementedError

    def geodesic_difference(self, x1, x0):
        """
        Compute and return the geodesic difference between 2 native states.
        This is a generalization of the Euclidean operation ``x1 - x0``.

        :param x1: native state.
        :param x0: native state.
        :return: Geodesic difference between native states ``x1`` and ``x2``.
        """
        return x1 - x0  # Default to Euclidean behavior.

    @abstractmethod
    def mean2pv(self, x):
        """
        Compute and return PV state from native state. Useful for combining
        state estimates of different types in IMM (Interacting Multiple Model)
        filtering.

        :param x: native state estimate mean.
        :return: PV state estimate mean.
        """
        raise NotImplementedError

    @abstractmethod
    def cov2pv(self, P):
        """
        Compute and return PV covariance from native covariance. Useful for
        combining state estimates of different types in IMM (Interacting
        Multiple Model) filtering.

        :param P: native state estimate covariance.
        :return: PV state estimate covariance.
        """
        raise NotImplementedError

    @abstractmethod
    def process_noise_cov(self, dt=0.0):
        """
        Compute and return process noise covariance (Q).

        :param dt: time interval to integrate over.
        :return: Read-only covariance (Q). For a DifferentiableDynamicModel, this is
            the covariance of the native state ``x`` resulting from stochastic
            integration (for use with EKF). Otherwise, it is the covariance
            directly of the process noise parameters (for use with UKF).
        """
        raise NotImplementedError

    def process_noise_dist(self, dt=0.0):
        """
        Return a distribution object of state displacement from the process noise
        distribution over a time interval.

        :param dt: time interval that process noise accumulates over.
        :return: :class:`~pyro.distributions.torch.MultivariateNormal`.
        """
        Q = self.process_noise_cov(dt)
        return dist.MultivariateNormal(
            torch.zeros(Q.shape[-1], dtype=Q.dtype, device=Q.device), Q
        )


class DifferentiableDynamicModel(DynamicModel):
    """
    DynamicModel for which state transition Jacobians can be efficiently
    calculated, usu. analytically or by automatic differentiation.
    """

    @abstractmethod
    def jacobian(self, dt):
        """
        Compute and return native state transition Jacobian (F) over time
        interval ``dt``.

        :param  dt: time interval to integrate over.
        :return: Read-only Jacobian (F) of integration map (f).
        """
        raise NotImplementedError


class Ncp(DifferentiableDynamicModel):
    """
    NCP (Nearly-Constant Position) dynamic model. May be subclassed, e.g., with
    CWNV (Continuous White Noise Velocity) or DWNV (Discrete White Noise
    Velocity).

    :param dimension: native state dimension.
    :param sv2: variance of velocity. Usually chosen so that the standard
          deviation is roughly half of the max velocity one would ever expect
          to observe.
    """

    def __init__(self, dimension, sv2):
        dimension_pv = 2 * dimension
        super().__init__(dimension, dimension_pv, num_process_noise_parameters=1)
        if not isinstance(sv2, torch.Tensor):
            sv2 = torch.tensor(sv2)
        self.sv2 = Parameter(sv2)
        self._F_cache = eye_like(sv2, dimension)  # State transition matrix cache
        self._Q_cache = {}  # Process noise cov cache

    def forward(self, x, dt, do_normalization=True):
        """
        Integrate native state ``x`` over time interval ``dt``.

        :param x: current native state. If the DynamicModel is non-differentiable,
              be sure to handle the case of ``x`` being augmented with process
              noise parameters.
        :param dt: time interval to integrate over.
            do_normalization: whether to perform normalization on output, e.g.,
            mod'ing angles into an interval. Has no effect for this subclass.
        :return: Native state x integrated dt into the future.
        """
        return x

    def mean2pv(self, x):
        """
        Compute and return PV state from native state. Useful for combining
        state estimates of different types in IMM (Interacting Multiple Model)
        filtering.

        :param x: native state estimate mean.
        :return: PV state estimate mean.
        """
        with torch.no_grad():
            x_pv = torch.zeros(2 * self._dimension, dtype=x.dtype, device=x.device)
            x_pv[: self._dimension] = x
        return x_pv

    def cov2pv(self, P):
        """
        Compute and return PV covariance from native covariance. Useful for
        combining state estimates of different types in IMM (Interacting
        Multiple Model) filtering.

        :param P: native state estimate covariance.
        :return: PV state estimate covariance.
        """
        d = 2 * self._dimension
        with torch.no_grad():
            P_pv = torch.zeros(d, d, dtype=P.dtype, device=P.device)
            P_pv[: self._dimension, : self._dimension] = P
        return P_pv

    def jacobian(self, dt):
        """
        Compute and return cached native state transition Jacobian (F) over
        time interval ``dt``.

        :param dt: time interval to integrate over.
        :return: Read-only Jacobian (F) of integration map (f).
        """
        return self._F_cache

    @abstractmethod
    def process_noise_cov(self, dt=0.0):
        """
        Compute and return cached process noise covariance (Q).

        :param dt: time interval to integrate over.
        :return: Read-only covariance (Q) of the native state ``x`` resulting from
            stochastic integration (for use with EKF).
        """
        raise NotImplementedError


class Ncv(DifferentiableDynamicModel):
    """
    NCV (Nearly-Constant Velocity) dynamic model. May be subclassed, e.g., with
    CWNA (Continuous White Noise Acceleration) or DWNA (Discrete White Noise
    Acceleration).

    :param dimension: native state dimension.
    :param sa2: variance of acceleration. Usually chosen so that the standard
          deviation is roughly half of the max acceleration one would ever
          expect to observe.
    """

    def __init__(self, dimension, sa2):
        dimension_pv = dimension
        super().__init__(dimension, dimension_pv, num_process_noise_parameters=1)
        if not isinstance(sa2, torch.Tensor):
            sa2 = torch.tensor(sa2)
        self.sa2 = Parameter(sa2)
        self._F_cache = {}  # State transition matrix cache
        self._Q_cache = {}  # Process noise cov cache

    def forward(self, x, dt, do_normalization=True):
        """
        Integrate native state ``x`` over time interval ``dt``.

        :param x: current native state. If the DynamicModel is non-differentiable,
              be sure to handle the case of ``x`` being augmented with process
              noise parameters.
        :param dt: time interval to integrate over.
        :param do_normalization: whether to perform normalization on output, e.g.,
              mod'ing angles into an interval. Has no effect for this subclass.

        :return: Native state x integrated dt into the future.
        """
        F = self.jacobian(dt)
        return F.mm(x.unsqueeze(1)).squeeze(1)

    def mean2pv(self, x):
        """
        Compute and return PV state from native state. Useful for combining
        state estimates of different types in IMM (Interacting Multiple Model)
        filtering.

        :param x: native state estimate mean.
        :return: PV state estimate mean.
        """
        return x

    def cov2pv(self, P):
        """
        Compute and return PV covariance from native covariance. Useful for
        combining state estimates of different types in IMM (Interacting
        Multiple Model) filtering.

        :param P: native state estimate covariance.
        :return: PV state estimate covariance.
        """
        return P

    def jacobian(self, dt):
        """
        Compute and return cached native state transition Jacobian (F) over
        time interval ``dt``.

        :param dt: time interval to integrate over.
        :return: Read-only Jacobian (F) of integration map (f).
        """
        if dt not in self._F_cache:
            d = self._dimension
            with torch.no_grad():
                F = eye_like(self.sa2, d)
                F[: d // 2, d // 2 :] = dt * eye_like(self.sa2, d // 2)
            self._F_cache[dt] = F

        return self._F_cache[dt]

    @abstractmethod
    def process_noise_cov(self, dt=0.0):
        """
        Compute and return cached process noise covariance (Q).

        :param dt: time interval to integrate over.
        :return: Read-only covariance (Q) of the native state ``x`` resulting from
            stochastic integration (for use with EKF).
        """
        raise NotImplementedError


class NcpContinuous(Ncp):
    """
    NCP (Nearly-Constant Position) dynamic model with CWNV (Continuous White
    Noise Velocity).

    References:
        "Estimation with Applications to Tracking and Navigation" by Y. Bar-
        Shalom et al, 2001, p.269.

    :param dimension: native state dimension.
    :param sv2: variance of velocity. Usually chosen so that the standard
          deviation is roughly half of the max velocity one would ever expect
          to observe.
    """

    def process_noise_cov(self, dt=0.0):
        """
        Compute and return cached process noise covariance (Q).

        :param dt: time interval to integrate over.
        :return: Read-only covariance (Q) of the native state ``x`` resulting from
            stochastic integration (for use with EKF).
        """
        if dt not in self._Q_cache:
            # q: continuous-time process noise intensity with units
            #   length^2/time (m^2/s). Choose ``q`` so that changes in position,
            #   over a sampling period ``dt``, are roughly ``sqrt(q*dt)``.
            q = self.sv2 * dt
            Q = q * dt * eye_like(self.sv2, self._dimension)
            self._Q_cache[dt] = Q

        return self._Q_cache[dt]


class NcvContinuous(Ncv):
    """
    NCV (Nearly-Constant Velocity) dynamic model with CWNA (Continuous White
    Noise Acceleration).

    References:
        "Estimation with Applications to Tracking and Navigation" by Y. Bar-
        Shalom et al, 2001, p.269.

    :param dimension: native state dimension.
    :param sa2: variance of acceleration. Usually chosen so that the standard
          deviation is roughly half of the max acceleration one would ever
          expect to observe.
    """

    def process_noise_cov(self, dt=0.0):
        """
        Compute and return cached process noise covariance (Q).

        :param dt: time interval to integrate over.

        :return: Read-only covariance (Q) of the native state ``x`` resulting from
            stochastic integration (for use with EKF).
        """
        if dt not in self._Q_cache:
            with torch.no_grad():
                d = self._dimension
                dt2 = dt * dt
                dt3 = dt2 * dt
                Q = torch.zeros(d, d, dtype=self.sa2.dtype, device=self.sa2.device)
                eye = eye_like(self.sa2, d // 2)
                Q[: d // 2, : d // 2] = dt3 * eye / 3.0
                Q[: d // 2, d // 2 :] = dt2 * eye / 2.0
                Q[d // 2 :, : d // 2] = dt2 * eye / 2.0
                Q[d // 2 :, d // 2 :] = dt * eye
            # sa2 * dt is an intensity factor that changes in velocity
            # over a sampling period ``dt``, ideally should be ~``sqrt(q*dt)``.
            Q = Q * (self.sa2 * dt)
            self._Q_cache[dt] = Q

        return self._Q_cache[dt]


class NcpDiscrete(Ncp):
    """
    NCP (Nearly-Constant Position) dynamic model with DWNV (Discrete White
    Noise Velocity).

    :param dimension: native state dimension.
    :param sv2: variance of velocity. Usually chosen so that the standard
          deviation is roughly half of the max velocity one would ever expect
          to observe.

    References:
        "Estimation with Applications to Tracking and Navigation" by Y. Bar-
        Shalom et al, 2001, p.273.
    """

    def process_noise_cov(self, dt=0.0):
        """
        Compute and return cached process noise covariance (Q).

        :param dt: time interval to integrate over.
        :return: Read-only covariance (Q) of the native state `x` resulting from
            stochastic integration (for use with EKF).
        """
        if dt not in self._Q_cache:
            Q = self.sv2 * dt * dt * eye_like(self.sv2, self._dimension)
            self._Q_cache[dt] = Q

        return self._Q_cache[dt]


class NcvDiscrete(Ncv):
    """
    NCV (Nearly-Constant Velocity) dynamic model with DWNA (Discrete White
    Noise Acceleration).

    :param dimension: native state dimension.
    :param sa2: variance of acceleration. Usually chosen so that the standard
          deviation is roughly half of the max acceleration one would ever
          expect to observe.

    References:
        "Estimation with Applications to Tracking and Navigation" by Y. Bar-
        Shalom et al, 2001, p.273.
    """

    def process_noise_cov(self, dt=0.0):
        """
        Compute and return cached process noise covariance (Q).

        :param dt: time interval to integrate over.
        :return: Read-only covariance (Q) of the native state `x` resulting from
            stochastic integration (for use with EKF). (Note that this Q, modulo
            numerical error, has rank `dimension/2`. So, it is only positive
            semi-definite.)
        """
        if dt not in self._Q_cache:
            with torch.no_grad():
                d = self._dimension
                dt2 = dt * dt
                dt3 = dt2 * dt
                dt4 = dt2 * dt2
                Q = torch.zeros(d, d, dtype=self.sa2.dtype, device=self.sa2.device)
                Q[: d // 2, : d // 2] = 0.25 * dt4 * eye_like(self.sa2, d // 2)
                Q[: d // 2, d // 2 :] = 0.5 * dt3 * eye_like(self.sa2, d // 2)
                Q[d // 2 :, : d // 2] = 0.5 * dt3 * eye_like(self.sa2, d // 2)
                Q[d // 2 :, d // 2 :] = dt2 * eye_like(self.sa2, d // 2)
            Q = Q * self.sa2
            self._Q_cache[dt] = Q

        return self._Q_cache[dt]



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

from abc import ABC, abstractmethod

import torch
import torch.nn

from .torch import TransformedDistribution
from .torch_transform import ComposeTransformModule


class ConditionalDistribution(ABC):
    @abstractmethod
    def condition(self, context):
        """:rtype: torch.distributions.Distribution"""
        raise NotImplementedError


class ConditionalTransform(ABC):
    @abstractmethod
    def condition(self, context):
        """:rtype: torch.distributions.Transform"""
        raise NotImplementedError


class ConditionalTransformModule(ConditionalTransform, torch.nn.Module):
    """
    Conditional transforms with learnable parameters such as normalizing flows should inherit from this class rather
    than :class:`~pyro.distributions.conditional.ConditionalTransform` so they are also a subclass of
    :class:`~torch.nn.Module` and inherit all the useful methods of that class.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __hash__(self):
        return super().__hash__()

    @property
    def inv(self) -> "ConditionalTransformModule":
        return _ConditionalInverseTransformModule(self)


class _ConditionalInverseTransformModule(ConditionalTransformModule):
    def __init__(self, transform: ConditionalTransform):
        super().__init__()
        self._transform = transform

    @property
    def inv(self) -> ConditionalTransform:
        return self._transform

    def condition(self, context: torch.Tensor):
        return self._transform.condition(context).inv


class ConditionalComposeTransformModule(
    ConditionalTransformModule, torch.nn.ModuleList
):
    """
    Conditional analogue of :class:`~pyro.distributions.torch_transform.ComposeTransformModule` .

    Useful as a base class for specifying complicated conditional distributions::

        >>> class ConditionalFlowStack(dist.conditional.ConditionalComposeTransformModule):
        ...     def __init__(self, input_dim, context_dim, hidden_dims, num_flows):
        ...         super().__init__([
        ...             dist.transforms.conditional_planar(input_dim, context_dim, hidden_dims)
        ...             for _ in range(num_flows)
        ...         ], cache_size=1)

        >>> cond_dist = dist.conditional.ConditionalTransformedDistribution(
        ...     dist.Normal(torch.zeros(3), torch.ones(3)).to_event(1),
        ...     [ConditionalFlowStack(3, 2, [8, 8], num_flows=4).inv]
        ... )

        >>> context = torch.rand(10, 2)
        >>> data = torch.rand(10, 3)
        >>> nll = -cond_dist.condition(context).log_prob(data)
    """

    def __init__(self, transforms, cache_size: int = 0):
        self.transforms = [
            (
                ConstantConditionalTransform(t)
                if not isinstance(t, ConditionalTransform)
                else t
            )
            for t in transforms
        ]
        super().__init__()
        if cache_size not in {0, 1}:
            raise ValueError("cache_size must be 0 or 1")
        self._cache_size = cache_size
        # for parameter storage
        for t in transforms:
            if isinstance(t, torch.nn.Module):
                self.append(t)

    def condition(self, context: torch.Tensor) -> ComposeTransformModule:
        return ComposeTransformModule(
            [t.condition(context) for t in self.transforms]
        ).with_cache(self._cache_size)


class ConstantConditionalDistribution(ConditionalDistribution):
    def __init__(self, base_dist):
        assert isinstance(base_dist, torch.distributions.Distribution)
        self.base_dist = base_dist

    def condition(self, context):
        return self.base_dist


class ConstantConditionalTransform(ConditionalTransform):
    def __init__(self, transform):
        assert isinstance(transform, torch.distributions.Transform)
        self.transform = transform

    def condition(self, context):
        return self.transform

    def clear_cache(self):
        self.transform.clear_cache()


class ConditionalTransformedDistribution(ConditionalDistribution):
    def __init__(self, base_dist, transforms):
        self.base_dist = (
            base_dist
            if isinstance(base_dist, ConditionalDistribution)
            else ConstantConditionalDistribution(base_dist)
        )
        self.transforms = [
            (
                t
                if isinstance(t, ConditionalTransform)
                else ConstantConditionalTransform(t)
            )
            for t in transforms
        ]

    def condition(self, context):
        base_dist = self.base_dist.condition(context)
        transforms = [t.condition(context) for t in self.transforms]
        return TransformedDistribution(base_dist, transforms)

    def clear_cache(self):
        pass



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import functools
import weakref
from collections import namedtuple

import torch
from torch.distributions import constraints

from pyro.distributions.util import broadcast_shape, is_validation_enabled
from pyro.ops.special import safe_log

from .torch_distribution import TorchDistribution


class CoalescentTimesConstraint(constraints.Constraint):
    def __init__(self, leaf_times, *, ordered=True):
        self.leaf_times = leaf_times
        self.ordered = ordered

    def check(self, value):
        # There must always at least one lineage.
        coal_times = value
        phylogeny = _make_phylogeny(self.leaf_times, coal_times)
        at_least_one_lineage = (phylogeny.lineages > 0).all(dim=-1)
        if not self.ordered:
            return at_least_one_lineage

        # Inputs must be ordered.
        ordered = (value[..., :-1] <= value[..., 1:]).all(dim=-1)
        return ordered & at_least_one_lineage


class CoalescentTimes(TorchDistribution):
    """
    Distribution over sorted coalescent times given irregular sampled
    ``leaf_times`` and constant population size.

    Sample values will be **sorted** sets of binary coalescent times. Each
    sample ``value`` will have cardinality ``value.size(-1) =
    leaf_times.size(-1) - 1``, so that phylogenies are complete binary trees.
    This distribution can thus be batched over multiple samples of phylogenies
    given fixed (number of) leaf times, e.g. over phylogeny samples from BEAST
    or MrBayes.

    **References**

    [1] J.F.C. Kingman (1982)
        "On the Genealogy of Large Populations"
        Journal of Applied Probability
    [2] J.F.C. Kingman (1982)
        "The Coalescent"
        Stochastic Processes and their Applications

    :param torch.Tensor leaf_times: Vector of times of sampling events, i.e.
        leaf nodes in the phylogeny. These can be arbitrary real numbers with
        arbitrary order and duplicates.
    :param torch.Tensor rate: Base coalescent rate (pairwise rate of
        coalescence) under a constant population size model. Defaults to 1.
    """

    arg_constraints = {"leaf_times": constraints.real, "rate": constraints.positive}

    def __init__(self, leaf_times, rate=1.0, *, validate_args=None):
        rate = torch.as_tensor(rate, dtype=leaf_times.dtype, device=leaf_times.device)
        batch_shape = broadcast_shape(rate.shape, leaf_times.shape[:-1])
        event_shape = (leaf_times.size(-1) - 1,)
        self.leaf_times = leaf_times
        self.rate = rate
        super().__init__(batch_shape, event_shape, validate_args=validate_args)

    @constraints.dependent_property
    def support(self):
        return CoalescentTimesConstraint(self.leaf_times)

    def log_prob(self, value):
        if self._validate_args:
            self._validate_sample(value)
        coal_times = value
        phylogeny = _make_phylogeny(self.leaf_times, coal_times)

        # The coalescent process is like a Poisson process with rate binomial
        # in the number of lineages, which changes at each event.
        binomial = phylogeny.binomial[..., :-1]
        interval = phylogeny.times[..., :-1] - phylogeny.times[..., 1:]
        log_prob = self.rate.log() * coal_times.size(-1) - self.rate * (
            binomial * interval
        ).sum(-1)

        # Scaling by those rates and accounting for log|jacobian|, the density
        # is that of a collection of independent Exponential intervals.
        log_abs_det_jacobian = phylogeny.coal_binomial.log().sum(-1).neg()
        return log_prob - log_abs_det_jacobian

    def sample(self, sample_shape=torch.Size()):
        shape = self._extended_shape(sample_shape)[:-1]
        leaf_times = self.leaf_times.expand(shape + (-1,))
        return _sample_coalescent_times(leaf_times)


class CoalescentTimesWithRate(TorchDistribution):
    """
    Distribution over coalescent times given irregular sampled ``leaf_times``
    and piecewise constant coalescent rates defined on a regular time grid.

    This assumes a piecewise constant base coalescent rate specified on time
    intervals ``(-inf,1]``, ``[1,2]``, ..., ``[T-1,inf)``,  where ``T =
    rate_grid.size(-1)``. Leaves may be sampled at arbitrary real times, but
    are commonly sampled in the interval ``[0, T]``.

    Sample values will be sorted sets of binary coalescent times. Each sample
    ``value`` will have cardinality ``value.size(-1) = leaf_times.size(-1) -
    1``, so that phylogenies are complete binary trees.  This distribution can
    thus be batched over multiple samples of phylogenies given fixed (number
    of) leaf times, e.g. over phylogeny samples from BEAST or MrBayes.

    This distribution implements :meth:`log_prob` but not ``.sample()``.

    See also :class:`~pyro.distributions.CoalescentRateLikelihood`.

    **References**

    [1] J.F.C. Kingman (1982)
        "On the Genealogy of Large Populations"
        Journal of Applied Probability
    [2] J.F.C. Kingman (1982)
        "The Coalescent"
        Stochastic Processes and their Applications
    [3] A. Popinga, T. Vaughan, T. Statler, A.J. Drummond (2014)
        "Inferring epidemiological dynamics with Bayesian coalescent inference:
        The merits of deterministic and stochastic models"
        https://arxiv.org/pdf/1407.1792.pdf

    :param torch.Tensor leaf_times: Tensor of times of sampling events, i.e.
        leaf nodes in the phylogeny. These can be arbitrary real numbers with
        arbitrary order and duplicates.
    :param torch.Tensor rate_grid: Tensor of base coalescent rates (pairwise
        rate of coalescence). For example in a simple SIR model this might be
        ``beta S / I``. The rightmost dimension is time, and this tensor
        represents a (batch of) rates that are piecewise constant in time.
    """

    arg_constraints = {
        "leaf_times": constraints.real,
        "rate_grid": constraints.positive,
    }

    def __init__(self, leaf_times, rate_grid, *, validate_args=None):
        batch_shape = broadcast_shape(leaf_times.shape[:-1], rate_grid.shape[:-1])
        event_shape = (leaf_times.size(-1) - 1,)
        self.leaf_times = leaf_times
        self.rate_grid = rate_grid
        super().__init__(batch_shape, event_shape, validate_args=validate_args)

    @constraints.dependent_property
    def support(self):
        return CoalescentTimesConstraint(self.leaf_times)

    @property
    def duration(self):
        return self.rate_grid.size(-1)

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(CoalescentTimesWithRate, _instance)
        new.leaf_times = self.leaf_times
        new.rate_grid = self.rate_grid
        super(CoalescentTimesWithRate, new).__init__(
            batch_shape, self.event_shape, validate_args=False
        )
        new._validate_args = self.__dict__.get("_validate_args")
        return new

    def log_prob(self, value):
        """
        Computes likelihood as in equations 7-8 of [3].

        This has time complexity ``O(T + S N log(N))`` where ``T`` is the
        number of time steps, ``N`` is the number of leaves, and ``S =
        sample_shape.numel()`` is the number of samples of ``value``.

        :param torch.Tensor value: A tensor of coalescent times. These denote
            sets of size ``leaf_times.size(-1) - 1`` along the trailing
            dimension and should be sorted along that dimension.
        :returns: Likelihood ``p(coal_times | leaf_times, rate_grid)``
        :rtype: torch.Tensor
        """
        if self._validate_args:
            self._validate_sample(value)
        coal_times = value
        phylogeny = _make_phylogeny(self.leaf_times, coal_times)

        # Compute survival factors for closed intervals.
        cumsum = self.rate_grid.cumsum(-1)
        cumsum = torch.nn.functional.pad(cumsum, (1, 0), value=0)
        integral = _interpolate_gather(
            cumsum, phylogeny.times[..., 1:]
        )  # ignore the final lonely leaf
        integral = integral[..., :-1] - integral[..., 1:]
        integral = integral.clamp(min=torch.finfo(integral.dtype).tiny)  # avoid nan
        log_prob = -(phylogeny.binomial[..., 1:-1] * integral).sum(-1)

        # Compute density of coalescent events.
        i = coal_times.floor().clamp(min=0, max=self.duration - 1).long()
        rates = phylogeny.coal_binomial * _gather(self.rate_grid, -1, i)
        log_prob = log_prob + safe_log(rates).sum(-1)

        batch_shape = broadcast_shape(self.batch_shape, value.shape[:-1])
        log_prob = log_prob.expand(batch_shape)
        return log_prob


class CoalescentRateLikelihood:
    """
    EXPERIMENTAL This is not a :class:`~pyro.distributions.Distribution`, but
    acts as a transposed version of :class:`CoalescentTimesWithRate` making the
    elements of ``rate_grid`` independent and thus compatible with ``plate``
    and ``poutine.markov``. For non-batched inputs the following are all
    equivalent likelihoods::

        # Version 1.
        pyro.sample("coalescent",
                    CoalescentTimesWithRate(leaf_times, rate_grid),
                    obs=coal_times)

        # Version 2. using pyro.plate
        likelihood = CoalescentRateLikelihood(leaf_times, coal_times, len(rate_grid))
        with pyro.plate("time", len(rate_grid)):
            pyro.factor("coalescent", likelihood(rate_grid))

        # Version 3. using pyro.markov
        likelihood = CoalescentRateLikelihood(leaf_times, coal_times, len(rate_grid))
        for t in pyro.markov(range(len(rate_grid))):
            pyro.factor("coalescent_{}".format(t), likelihood(rate_grid[t], t))

    The third version is useful for e.g.
    :class:`~pyro.infer.smcfilter.SMCFilter` where ``rate_grid`` might be
    computed sequentially.

    :param torch.Tensor leaf_times: Tensor of times of sampling events, i.e.
        leaf nodes in the phylogeny. These can be arbitrary real numbers with
        arbitrary order and duplicates.
    :param torch.Tensor coal_times: A tensor of coalescent times. These denote
        sets of size ``leaf_times.size(-1) - 1`` along the trailing dimension
        and should be sorted along that dimension.
    :param int duration: Size of the rate grid, ``rate_grid.size(-1)``.
    """

    def __init__(self, leaf_times, coal_times, duration, *, validate_args=None):
        assert leaf_times.size(-1) == 1 + coal_times.size(-1)
        assert isinstance(duration, int) and duration >= 2
        if validate_args is True or validate_args is None and is_validation_enabled:
            constraint = CoalescentTimesConstraint(leaf_times, ordered=False)
            if not constraint.check(coal_times).all():
                raise ValueError("Invalid (leaf_times, coal_times)")

        phylogeny = _make_phylogeny(leaf_times, coal_times)
        batch_shape = phylogeny.times.shape[:-1]
        new_zeros = leaf_times.new_zeros
        new_ones = leaf_times.new_ones

        # Construct linear part from intervals of survival outside of [0,duration].
        times = phylogeny.times.clamp(max=0)
        intervals = times[..., 1:] - times[..., :-1]
        pre_linear = (phylogeny.binomial[..., :-1] * intervals).sum(-1, keepdim=True)
        times = phylogeny.times.clamp(min=duration)
        intervals = times[..., 1:] - times[..., :-1]
        post_linear = (phylogeny.binomial[..., :-1] * intervals).sum(-1, keepdim=True)
        self._linear = torch.cat(
            [
                pre_linear,
                new_zeros(pre_linear.shape[:-1] + (duration - 2,)),
                post_linear,
            ],
            dim=-1,
        )

        # Construct linear part from intervals of survival within [0, duration].
        times = phylogeny.times.clamp(min=0, max=duration)
        sparse_diff = phylogeny.binomial[..., :-1] - phylogeny.binomial[..., 1:]
        dense_diff = new_zeros(batch_shape + (1 + duration,))
        _interpolate_scatter_add_(dense_diff, times[..., 1:], sparse_diff)
        self._linear += dense_diff.flip([-1]).cumsum(-1)[..., :-1].flip([-1])

        # Construct const and log part from coalescent events.
        coal_index = coal_times.floor().clamp(min=0, max=duration - 1).long()
        self._const = new_zeros(batch_shape + (duration,))
        self._const.scatter_add_(-1, coal_index, phylogeny.coal_binomial.log())
        self._log = new_zeros(batch_shape + (duration,))
        self._log.scatter_add_(-1, coal_index, new_ones(coal_index.shape))

    def __call__(self, rate_grid, t=slice(None)):
        """
        Computes the likelihood of [1] equations 7-9 for one or all time
        points.

        **References**

        [1] A. Popinga, T. Vaughan, T. Statler, A.J. Drummond (2014)
            "Inferring epidemiological dynamics with Bayesian coalescent
            inference: The merits of deterministic and stochastic models"
            https://arxiv.org/pdf/1407.1792.pdf

        :param torch.Tensor rate_grid: Tensor of base coalescent rates
            (pairwise rate of coalescence). For example in a simple SIR model
            this might be ``beta S / I``. The rightmost dimension is time, and
            this tensor represents a (batch of) rates that are piecwise
            constant in time.
        :param time: Optional time index by which the input was sliced, as in
            ``rate_grid[..., t]`` This can be an integer for sequential models
            or ``slice(None)`` for vectorized models.
        :type time: int or slice
        :returns: Likelihood ``p(coal_times | leaf_times, rate_grid)``,
            or a part of that likelihood corresponding to a single time step.
        :rtype: torch.Tensor
        """
        const = self._const[..., t]
        linear = self._linear[..., t] * rate_grid
        log = (
            self._log[..., t]
            * rate_grid.clamp(min=torch.finfo(rate_grid.dtype).tiny).log()
        )
        return const + linear + log


def bio_phylo_to_times(tree, *, get_time=None):
    """
    Extracts coalescent summary statistics from a phylogeny, suitable for use
    with :class:`~pyro.distributions.CoalescentRateLikelihood`.

    :param Bio.Phylo.BaseTree.Clade tree: A phylogenetic tree.
    :param callable get_time: Optional function to extract the time point of
        each sub-:class:`~Bio.Phylo.BaseTree.Clade`. If absent, times will be
        computed by cumulative `.branch_length`.
    :returns: A pair of :class:`~torch.Tensor` s ``(leaf_times, coal_times)``
        where ``leaf_times`` are times of sampling events (leaf nodes in the
        phylogenetic tree) and ``coal_times`` are times of coalescences (leaf
        nodes in the phylogenetic binary tree).
    :rtype: tuple
    """
    if get_time is None:
        # Compute time as cumulative branch length.
        def get_branch_length(clade):
            branch_length = clade.branch_length
            return 1.0 if branch_length is None else branch_length

        times = {tree.root: get_branch_length(tree.root)}

    leaf_times = []
    coal_times = []
    for clade in tree.find_clades():
        if get_time is None:
            time = times[clade]
            for child in clade:
                times[child] = time + get_branch_length(child)
        else:
            time = get_time(clade)

        num_children = len(clade)
        if num_children == 0:
            leaf_times.append(time)
        else:
            # Pyro expects binary coalescent events, so we split n-ary events
            # into n-1 separate binary events.
            for _ in range(num_children - 1):
                coal_times.append(time)
    assert len(leaf_times) == 1 + len(coal_times)

    leaf_times = torch.tensor(leaf_times)
    coal_times = torch.tensor(coal_times)
    return leaf_times, coal_times


def _gather(tensor, dim, index):
    """
    Like :func:`torch.gather` but broadcasts.
    """
    if dim != -1:
        raise NotImplementedError
    shape = broadcast_shape(tensor.shape[:-1], index.shape[:-1]) + (-1,)
    tensor = tensor.expand(shape)
    index = index.expand(shape)
    return tensor.gather(dim, index)


def _interpolate_gather(array, x):
    """
    Like ``torch.gather(-1, array, x)`` but continuously indexes into the
    rightmost dim of an array, linearly interpolating between array values.
    """
    with torch.no_grad():
        x0 = x.floor().clamp(min=0, max=array.size(-1) - 2)
        x1 = x0 + 1
    f0 = _gather(array, -1, x0.long())
    f1 = _gather(array, -1, x1.long())
    return f0 * (x1 - x) + f1 * (x - x0)


def _interpolate_scatter_add_(dst, x, src):
    """
    Like ``dst.scatter_add_(-1, x, src)`` but continuously index into the
    rightmost dim of an array, linearly interpolating between array values.
    """
    with torch.no_grad():
        x0 = x.floor().clamp(min=0, max=dst.size(-1) - 2)
        x1 = x0 + 1
    dst.scatter_add_(-1, x0.long(), src * (x1 - x))
    dst.scatter_add_(-1, x1.long(), src * (x - x0))
    return dst


def _weak_memoize(fn):
    cache = {}

    @functools.wraps(fn)
    def memoized_fn(*args):
        key = tuple(map(id, args))

        # Allow cache hit only when tensors have not since been mutated.
        version = tuple(arg._version for arg in args)
        if key in cache:
            old_version, result = cache[key]
            if old_version == version:
                return result

        result = fn(*args)
        cache[key] = version, result
        for arg in args:
            weakref.finalize(arg, cache.pop, key, None)
        return result

    return memoized_fn


# This helper data structure has only timing information.
_Phylogeny = namedtuple(
    "_Phylogeny",
    (
        "times",
        "signs",
        "lineages",
        "binomial",
        "coal_binomial",
    ),
)


@_weak_memoize
@torch.no_grad()
def _make_phylogeny(leaf_times, coal_times):
    assert leaf_times.size(-1) == 1 + coal_times.size(-1)

    # Expand shapes to match.
    N = leaf_times.size(-1)
    batch_shape = broadcast_shape(leaf_times.shape[:-1], coal_times.shape[:-1])
    if leaf_times.shape[:-1] != batch_shape:
        leaf_times = leaf_times.expand(batch_shape + (N,))
    if coal_times.shape[:-1] != batch_shape:
        coal_times = coal_times.expand(batch_shape + (N - 1,))

    # Combine N sampling events (leaf_times) plus N-1 coalescent events
    # (coal_times) into a pair (times, signs) of arrays of length 2N-1, where
    # leaf sample sign is +1 and coalescent sign is -1.
    times = torch.cat([coal_times, leaf_times], dim=-1)
    signs = torch.linspace(
        1.5 - N, N - 0.5, 2 * N - 1
    ).sign()  # e.g. [-1, -1, +1, +1, +1]

    # Sort the events reverse-ordered in time, i.e. latest to earliest.
    times, index = times.sort(dim=-1, descending=True)
    signs = signs[index]
    inv_index = index.new_empty(index.shape)
    inv_index.scatter_(-1, index, torch.arange(2 * N - 1).expand_as(index))

    # Compute the number n of lineages preceding each event, then the binomial
    # coefficients that will multiply the base coalescence rate.
    lineages = signs.cumsum(-1)
    binomial = lineages * (lineages - 1) / 2

    # Compute the binomial coefficient following each coalescent event.
    coal_index = inv_index[..., : N - 1]
    coal_binomial = binomial.gather(-1, coal_index - 1)

    return _Phylogeny(times, signs, lineages, binomial, coal_binomial)


def _sample_coalescent_times(leaf_times):
    leaf_times = leaf_times.detach()
    proto = leaf_times
    N = leaf_times.size(-1)
    batch_shape = leaf_times.shape[:-1]

    # We don't bother to implement a version that vectorizes over batches;
    # instead we simply sequentially sample and stack.
    if batch_shape:
        flat_leaf_times = leaf_times.reshape(-1, N)
        flat_coal_times = torch.stack(
            list(map(_sample_coalescent_times, flat_leaf_times))
        )
        return flat_coal_times.reshape(batch_shape + (N - 1,))
    assert leaf_times.shape == (N,)

    # Sequentially sample coalescent events from latest to earliest.
    leaf_times = leaf_times.sort(dim=-1, descending=True).values.tolist()
    coal_times = []
    # Start with the minimum of two active leaves.
    leaf = 1
    t = leaf_times[leaf]
    active = 2
    binomial = active * (active - 1) / 2
    for u in proto.new_empty(N - 1).exponential_().tolist():
        while leaf + 1 < N and u > (t - leaf_times[leaf + 1]) * binomial:
            # Move past the next leaf.
            leaf += 1
            u -= (t - leaf_times[leaf]) * binomial
            t = leaf_times[leaf]
            active += 1
            binomial = active * (active - 1) / 2
        # Add a coalescent event.
        t -= u / binomial
        active -= 1
        binomial = active * (active - 1) / 2
        coal_times.append(t)
    coal_times.reverse()

    return proto.new_tensor(coal_times)
